import logging

import numpy as np 

from scipy.interpolate import interp1d
from astropy.table import Table


def stack_spectra_bootstrap(survey, 
                              data_column = None, 
                              variance_column = None, 
                              velocity = None, 
                              n_boot = None, 
                              ci = None, 
                              estimator = None, 
                              set_name = None):
    """
    Stack all spectra in provided SkySurvey Table
    
    Parameters
    ----------
    survey: 'whampy.skySurvey'
        input skySurvey
    data_column: 'str', optional, must be keyword
        Name of data column, default of "DATA"
    variance_column: 'str', optional, must be keyword
        Name od variance column, defaut of "VARIANCE"
    velocity: 'np.array, list', optional, must be keyword
        velocity to interpolate data to for stacking
        defaults to np.round of first row "VELOCITY"
    n_boot: 'int', optional, must be keyword
        number of bootstrap samples for each spectrum to stack
        default of 1000
    ci: 'number', optional, must be keyword
        Size of confidence interval to return as errors
    estimator: 'callable', optional, must be keyword
        estimator method to use, defaults to `numpy.mean`
        must take axis keyword
    set_name: 'str', optional, must be keyword
        if provided, sets "NAME" column to this in returned stack
    """
    
    # Set Defaults
    if data_column is None:
        data_column = "DATA"
    if variance_column is None:
        variance_column = "VARIANCE"
    if velocity is None:
        velocity = survey[0]["VELOCITY"]
    if n_boot is None:
        n_boot = 10000
    if ci is None:
        ci = 95
    if estimator is None:
        estimator = np.nanmean
        

    # data_interp = np.zeros_like(survey[data_column])
    # for ell, (velocity_window, 
    #           data_window) in enumerate(zip(survey["VELOCITY"], 
    #                                         survey[data_column])):
    #     f = interp1d(velocity_window, 
    #                              data_window, 
    #                              fill_value= "extrapolate")
    #     data_interp[ell,:] = f(velocity)

        
    stacked = Table()
    stacked["VELOCITY"] = np.zeros((1, len(velocity)))
    stacked["DATA"] = np.zeros((1, len(velocity)))
    stacked["VARIANCE"] = np.zeros((1, len(velocity)))
    stacked["CI"] = np.zeros((1, 2, len(velocity)))

    data_length = len(survey[data_column][0])

    samples_data = []
    samples_inds = np.arange(len(survey), dtype = int)
    for ell in range(n_boot):
        samples_data_sub = []
        choice_size = len(survey)
        chosen_inds = np.random.choice(samples_inds, size = choice_size)
        # samples_data[ell].append(np.random.choice(survey[data_column][:,ell], size = choice_size))
        samples_data_sub.append(survey[data_column][chosen_inds, :])
        stack_samples_data_sub = np.vstack(samples_data_sub)
        samples_data.append(estimator(stack_samples_data_sub, axis = 0))

    stack_samples_data = np.stack(samples_data)
    
    stacked[0]["VELOCITY"] = velocity
    stacked[0]["DATA"] = estimator(stack_samples_data, 
                                axis = 0) * survey["DATA"].unit
    stacked[0]["VARIANCE"] = np.var(stack_samples_data, 
                                 axis = 0) * survey["VARIANCE"].unit
    stacked[0]["CI"] = np.nanpercentile(stack_samples_data, 
                                  (100-ci, ci), 
                                  axis = 0) * survey["DATA"].unit


    if set_name is not None:
        stacked["NAME"] = [set_name]
        
    return survey.__class__(from_table = stacked)

def combine_velocity_windows_bootstrap(obs_list, 
                                       survey_spectra = None,
                                       exp_times = None, 
                              data_column = None, 
                              variance_column = None, 
                              velocity = None, 
                              n_boot = None, 
                                       n_survey_sample = None,
                              ci = None, 
                              estimator = None, 
                              set_name = None, 
                                       vmin = None, 
                                       vmax = None, 
                                       v_res = None):
    """
    Parameters
    ----------
    obs_list: 'list'
        input SkySurveys as a list
    survey_spectra: `SkySurvey`, optional
        single spectrum from WHAM-SS
        obs variances will be used for random draws
    exp_times: 'list', optional
        list of exposure times for weighting
        NOT YET IMPLEMENTED
    data_column: 'str', optional, must be keyword
        Name of data column, default of "DATA"
    variance_column: 'str', optional, must be keyword
        Name od variance column, defaut of "VARIANCE"
    velocity: 'np.array, list', optional, must be keyword
        velocity to interpolate data to for stacking
    n_boot: 'int', optional, must be keyword
        number of bootstrap samples for each spectrum to stack
        default of 1000
    n_survey_sample: 'int', optional, must be keyword
        number of bootstrap samples for each spectrum to stack
        default of 1000
    ci: 'number', optional, must be keyword
        Size of confidence interval to return as errors
    estimator: 'callable', optional, must be keyword
        estimator method to use, defaults to `numpy.mean`
        must take axis keyword
    set_name: 'str', optional, must be keyword
        if provided, sets "NAME" column to this in returned stack
    vmin: `number`, optional, must be keyword
        min velocity to consider
    vmax: `number`, optional, must be keyword
        max velocity to consider
    v_res: `number` optional, must be keyword
        velocity resolution / channel width in km/s
    """
    
    if data_column is None:
        data_column = "DATA"
    if variance_column is None:
        variance_column = "VARIANCE"
    if n_boot is None:
        n_boot = 10000
    if ci is None:
        ci = 95
    if estimator is None:
        estimator = np.nanmean
        
    # Ensure obslist is not a single SkySurvey Object
    if hasattr(obs_list, "stack_spectra_bootstrap"):
        raise TypeError("obs_list must be SkySurvey's wrapped in a list!")
        
        
    # Determine number of windows to merge
    n_windows = len(obs_list)
    
    # Determine highest and lowest velocities if needed
    if velocity is None:
        if vmin is None:
            all_vmins = [np.nanmin(obs["VELOCITY"]) for obs in obs_list]
            vmin = np.nanmin(all_vmins)

        if vmax is None:
            all_vmaxs = [np.nanmax(obs["VELOCITY"]) for obs in obs_list]
            vmax = np.nanmax(all_vmaxs)

        # Set Velocity Window
        if v_res is None:
            v_res = 1.5
        
        velocity = np.arange(vmin, vmax + v_res, v_res)

    data_length = len(velocity)
        
    # Interpolate spectra:
    interp_data = []
    for obs in obs_list:
        data_this_row = []
        for row in obs:
            data_interpolator = interp1d(row["VELOCITY"], row[data_column], 
                                         bounds_error = False, fill_value = np.nan)
            data_this_row.append([data_interpolator(velocity)])
        interp_data.append(np.vstack(data_this_row))
        
    if survey_spectra is not None:
        survey_data_interpolator = interp1d(survey_spectra["VELOCITY"], survey_spectra[data_column], 
                                            bounds_error = False, fill_value = np.nan)
        survey_data = survey_data_interpolator(velocity)
        survey_variance_interpolator = interp1d(survey_spectra["VELOCITY"], survey_spectra[variance_column], 
                                            bounds_error = False, fill_value = np.nan)
        survey_variance = survey_variance_interpolator(velocity)
        
        # Apply Random Perturbations
        
        if n_survey_sample is None:
            n_survey_sample = 1000
            
        data_this_row = []
        for ell in range(n_survey_sample):
            random_errors = np.sqrt(survey_variance) * 3 * np.random.randn(len(velocity))
            data_random_survey = survey_data + random_errors
            data_this_row.append(data_random_survey)
            
        interp_data.append(np.vstack(data_this_row))
    
    
    stacked = Table()
    stacked["VELOCITY"] = np.zeros((1, len(velocity)))
    stacked["DATA"] = np.zeros((1, len(velocity)))
    stacked["VARIANCE"] = np.zeros((1, len(velocity)))
    stacked["CI"] = np.zeros((1, 2, len(velocity)))
    

    samples_data = []
    for iteration in range(n_boot):
        samples_data_sub = []
        for data in interp_data:
            data_inds = np.arange(len(data), dtype = int)
            choice_size = len(data)
            chosen_inds = np.random.choice(data_inds, size = choice_size)
            samples_data_sub.append(data[chosen_inds])
       
        stack_samples_data_sub = np.vstack(samples_data_sub)
        x = estimator(stack_samples_data_sub, axis = 0)
        samples_data.append(x)
            
            

    stack_samples_data = np.stack(samples_data)

    stacked[0]["VELOCITY"] = velocity
    stacked[0]["DATA"] = estimator(stack_samples_data, 
                                axis = 0) * obs_list[0]["DATA"].unit
    stacked[0]["VARIANCE"] = np.nanvar(stack_samples_data, 
                                 axis = 0) * obs_list[0]["VARIANCE"].unit
    stacked[0]["CI"] = np.nanpercentile(stack_samples_data, 
                                  (100-ci, ci), 
                                  axis = 0) * obs_list[0]["DATA"].unit
    
    if set_name is not None:
        stacked["NAME"] = [set_name]
        
    return obs_list[0].__class__(from_table = stacked)
    



import pytest
import numpy as np
from ..skySurvey import SkySurvey
from ..skySurvey import directory

from ..scaleHeight import fit_scale_heights

from dustmaps.bayestar import BayestarWebQuery
# Set up the random number generator
np.random.seed(1234)

# Load global survey
survey = SkySurvey()
bayestar = BayestarWebQuery()

data, df, masks = survey.get_scale_height_data(track = 'SgN',
                                            return_pandas_dataframe = True, 
                                            longitude_mask_width = 3, 
                                            lrange = [30, 35])
data_dered, df_dered, masks_dered = survey.get_scale_height_data(track = 'SgN',
                                            return_pandas_dataframe = True, 
                                            longitude_mask_width = 3, 
                                            lrange = [30, 35], deredden = bayestar)

def test_basic():
    """
    Basic test
    """
    results = fit_scale_heights(data, masks, return_smoothed = True)
    results_2 = fit_scale_heights(data_dered, masks_dered, 
        return_smoothed = True, 
        fig_names = "test_fits", 
        smoothed_width = 5, 
        min_lat = 5, 
        max_lat = 35, 
        deredden = True)

    assert np.allclose(results["slopes_pos"], results_2["slopes_pos"], equal_nan = True)


import pytest
import matplotlib.pyplot as plt
import astropy.units as u
import numpy as np
from ..skySurvey import SkySurvey 
from unittest.mock import Mock

from ..skySurvey import directory

# Set up the random number generator.
np.random.seed(1234)

# Load survey
survey = SkySurvey()

BASELINE_DIR = 'baseline'

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_event_over():
    """
    Test click map with sample click event on SkySurvey over
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    click_map = survey.click_map(fig = fig, image_ax = ax, over_data = survey, share_yaxis = True)
    event = Mock()
    event.button = 1
    event.inaxes = ax
    event.xdata = 1
    event.ydata = -4
    click_map.on_click(event)
    return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_event_over_notsharey():
    """
    Test click map with sample click event on SkySurvey over
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    click_map = survey.click_map(fig = fig, image_ax = ax, over_data = survey, share_yaxis = False)
    event = Mock()
    event.button = 1
    event.inaxes = ax
    event.xdata = 1
    event.ydata = -4
    click_map.on_click(event)
    return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_str_click_event_over():
    """
    Test click map with sample click event on string load over
    """
    import os.path
    over_data = os.path.join(directory, "tests/test_data/test_cube.fits")
    fig = plt.figure()
    ax = fig.add_subplot(111)
    click_map = survey.click_map(fig = fig, image_ax = ax, over_data = over_data)
    event = Mock()
    event.button = 1
    event.inaxes = ax
    event.xdata = 1
    event.ydata = -4
    click_map.on_click(event)
    return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance=20)
def test_cube_click_event_over():
    """
    Test click map with sample click event on string load over
    """
    import os.path
    from spectral_cube import SpectralCube
    over_data_path = os.path.join(directory, "tests/test_data/test_cube.fits")
    over_data = SpectralCube.read(over_data_path)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    click_map = survey.click_map(fig = fig, image_ax = ax, over_data = over_data, 
        average_beam = True, radius = 0.5)
    event = Mock()
    event.button = 1
    event.inaxes = ax
    event.xdata = 1
    event.ydata = -4
    click_map.on_click(event)
    return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance=20)
def test_wcs_click_event_over():
    """
    Ensure click map works with wcs axes
    """
    import os.path
    from spectral_cube import SpectralCube
    over_data_path = os.path.join(directory, "tests/test_data/test_cube.fits")
    over_data = SpectralCube.read(over_data_path)
    fig = plt.figure()
    ax = fig.add_subplot(111, projection = over_data.wcs, slices = ['x', 'y', 0])
    ax.imshow(over_data[0,:,:].data)

    click_map = survey.click_map(fig = fig, image_ax = ax, over_data = over_data)
    x,y,_  = over_data.wcs.wcs_world2pix(1,-4,0,0)
    event = Mock()
    event.button = 1
    event.inaxes = ax
    event.xdata = x
    event.ydata = y
    click_map.on_click(event)
    return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance=20)
def test_cube_click_event_over_radius_deg():
    """
    Test click map with sample click event on string load over with radius quantity
    """
    import os.path
    from spectral_cube import SpectralCube
    over_data_path = os.path.join(directory, "tests/test_data/test_cube.fits")
    over_data = SpectralCube.read(over_data_path)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    click_map = survey.click_map(fig = fig, image_ax = ax, over_data = over_data, 
        average_beam = True, radius = 0.5*u.deg)
    event = Mock()
    event.button = 1
    event.inaxes = ax
    event.xdata = 1
    event.ydata = -4
    click_map.on_click(event)
    return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance=20)
def test_cube_click_event_over_radius_none():
    """
    Test click map with sample click event on string load over with radius not set
    """
    import os.path
    from spectral_cube import SpectralCube
    over_data_path = os.path.join(directory, "tests/test_data/test_cube.fits")
    over_data = SpectralCube.read(over_data_path)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    click_map = survey.click_map(fig = fig, image_ax = ax, over_data = over_data, 
        average_beam = True)
    event = Mock()
    event.button = 1
    event.inaxes = ax
    event.xdata = 1
    event.ydata = -4
    click_map.on_click(event)
    return plt.gcf()

def test_cube_over_readerror():
    """
    Test click map with over data read error
    """
    try:
        click_map = survey.click_map(over_data = "fakefilename.fits")
    except FileNotFoundError:
        assert True
    else:
        assert False





















import logging

from astropy import units as u
import numpy as np 
import matplotlib.pyplot as plt
from scipy import interpolate

from astropy.coordinates import SkyCoord
from astropy.coordinates import Angle

try:
    from extinction import fm07 as extinction_law
except ModuleNotFoundError:
    # Error handling
    pass
try:
    from dustmaps.marshall import MarshallQuery
    from dustmaps.bayestar import BayestarQuery
    from dustmaps.bayestar import BayestarWebQuery
except ModuleNotFoundError:
    # Error handling
    pass

from scipy.interpolate import interp1d
from scipy import stats

import seaborn as sns
pal = sns.color_palette("colorblind")

import pandas as pd


def kinematic_distance_flat(l, v, R_sun = 8.12 * u.kpc, v_sun = 220 * u.km/u.s):
    term1 = R_sun * np.cos(l*u.deg)
    r = R_sun * np.sin(l*u.deg) * v_sun / (v*u.km/u.s + v_sun * np.sin(l*u.deg))
    term2 = np.sqrt(r**2 - (R_sun *  np.sin(l*u.deg))**2)
    
    return term1 + term2, term1 - term2


def get_scale_height_data(data, track = None, deredden = False, 
                          return_pandas_dataframe = False, 
                          longitude_mask_width = None, 
                          step_size = None, R_sun = None, v_sun = None, 
                          closer = False, add_kinematic_distance = False, **kwargs):
    """
    Return data needed for scale height analysis
    
    Parameters
    ----------
    data: `skySurvey`
        WHAM skySurvey object of full sky (requires track keyword), or spiral arm section
    track: `str`, optional, must be keyword
        if provided, will apply skySurvey.get_spiral_slice for provided track
        if None, will check for track as an attribute of data
    deredden: `bool`, `dustmap`, optional, must be keyword
        if True, will apply dereddening using 3D dustmaps of Marshall et al. (2006)
        or can input a dustmap to query from using the dustmaps package
        default to `dustmaps.marshall.MarshallQuery`
        Warning: Currently only supports Marshall Dustmap
    return_pandas_dataframe: `bool`, optional, must be keyword
        if True, returns pandas dataframe with subset of data specific to scale height analysis
    longitude_mask_width: `number`, `u.Quantity`, optional, must be keyword
        if provided, returns list of masks splitting data into sky sections 
        of given width in degrees
    step_size: `number`, `u.Quantity`, optional, must be keyword
        if provided, sets step_size for longitude masks
        default to half width
    R_sun: `u.Quantity`, optional, must be keyword
        Sun Galactocentric Distance
    v_sun: `u.Quantity`, optional, must be keyword
        Sun rotation velocity
    add_kinematic_distance: `bool`, optional, must be keyword
        if True, adds in kinematic distances using a flat rotation curve 
        where no parallax ones available

    **kwargs: `dict`
        keywords passed to data.get_spiral_slice if track is provided
    """
    # Check Wrapping
    if "wrap_at_180" in kwargs:
        if kwargs["wrap_at_180"]:
            wrap_at = "180d"
        else:
            wrap_at = "360d"
    else:
        wrap_at = "360d"

    if add_kinematic_distance:
        if R_sun is None:
            R_sun = 8.127 * u.kpc
        if v_sun is None:
            v_sun = 220*u.km/u.s
        
    # Must have return_track
    try:
        test = np.all(kwargs["return_track"])
    except KeyError:
        kwargs["return_track"] = True
    finally: 
        if kwargs["return_track"] is False:
            kwargs["return_track"] = True
            logging.warning("keyword 'return_track' must be set to True!")
    
    # Get / ensure proper track is available
    if (track is None):
        if not hasattr(data, "lbv_RBD_track"):
            raise SyntaxError("No track provided - distance information is required")
    else:
        data, data.lbv_RBD_track = data.get_spiral_slice(track = track, **kwargs)
    
    if data.lbv_RBD_track.shape[1] < 3:
        raise ValueError("provided track does not have distance information!")

    if add_kinematic_distance:
        distance_is_none = data.lbv_RBD_track[:,-1] == 0.0
        if distance_is_none.sum() > 0:
            distances_1, distances_2  = kinematic_distance_flat(data.lbv_RBD_track[distance_is_none,0], 
                data.lbv_RBD_track[distance_is_none,1], R_sun = R_sun, v_sun = v_sun)

            if closer:
                neg_dist = distances_2 < 0.
                distances_2[neg_dist] = distances_1[neg_dist]
                data.lbv_RBD_track[distance_is_none,-1] = distances_2
            else:
                data.lbv_RBD_track[distance_is_none,-1] = distances_1



    
    # Setup dustmaps if needed
    if deredden.__class__ is bool:
        if deredden:
            deredden = MarshallQuery()
    elif not hasattr(deredden, "query"):
        raise TypeError("invaled dustmap provided - must provide a dustmap class that can query or set to \
                        True to set defualt dustmap to MarshallQuery.")
        
    data["tan(b)"] = np.tan(data["GAL-LAT"].data*u.deg)
    
    # Apply dereddening
    if not deredden.__class__ is bool:

        # Get all distances assuming plane parallel
        distance_interpolator = interpolate.interp1d(Angle(data.lbv_RBD_track[:,0]*u.deg).wrap_at(wrap_at), 
                                                     data.lbv_RBD_track[:,-1])
        distances = distance_interpolator(Angle(data["GAL-LON"].data*u.deg).wrap_at(wrap_at))
        coordinates = data.get_SkyCoord(distance = distances * u.kpc)


        if (deredden.__class__ is BayestarQuery) | (deredden.__class__ is BayestarWebQuery):
            Av_bayestar = 2.742 * deredden(coordinates)
            wave_ha = np.array([6562.8])
            A_V_to_A_ha = extinction_law(wave_ha, 1.)
            data["DISTANCE"] = distances * u.kpc
            data["Z"] = data["DISTANCE"] * data["tan(b)"]
            data["Av"] = Av_bayestar
            data["INTEN_DERED"] = data["INTEN"][:]
            data["INTEN_DERED"][~np.isnan(Av_bayestar)] = \
                    data["INTEN"][~np.isnan(Av_bayestar)] * 10**(0.4 * A_V_to_A_ha  * Av_bayestar[~np.isnan(Av_bayestar)])
        else:

            
            # Get A_Ks
            AKs = deredden(coordinates)
            wave_Ks = 2.17 *u.micron
            A_KS_to_A_v = 1. / extinction_law(np.array([wave_Ks.to(u.AA).value]), 1.)
            wave_ha = np.array([6562.8])
            A_V_to_A_ha = extinction_law(wave_ha, 1.)
            data["DISTANCE"] = distances * u.kpc
            data["Z"] = data["DISTANCE"] * data["tan(b)"]
            data["Av"] = A_KS_to_A_v * AKs
            data["INTEN_DERED"] = data["INTEN"][:]
            data["INTEN_DERED"][~np.isnan(AKs)] = \
                    data["INTEN"][~np.isnan(AKs)] * 10**(0.4 * A_V_to_A_ha * A_KS_to_A_v * AKs[~np.isnan(AKs)])
    
    if not longitude_mask_width is None:
        if not isinstance(longitude_mask_width, u.Quantity):
            longitude_mask_width *= u.deg
            logging.warning("No units provided for longitude_mask_width, assuming u.deg.")
        if step_size is None:
            step_size = longitude_mask_width / 2.
        elif not isinstance(step_size, u.Quantity):
            step_size *= u.deg
            logging.warning("No units provided for step_size, assuming u.deg.")

        # Construct masks
        wrapped_lon = Angle(data["GAL-LON"]).wrap_at(wrap_at)
        lon_range = np.min(wrapped_lon), np.max(wrapped_lon)
        n_steps = int(np.ceil(np.round(lon_range[1] - lon_range[0]) / step_size))
        lon_edge = np.linspace(lon_range[0], lon_range[1], n_steps)
        lon_edges = np.zeros((len(lon_edge)-1, 2)) * u.deg
        lon_edges[:,0] = lon_edge[:-1] 
        lon_edges[:,1] = lon_edge[:-1] + longitude_mask_width
        masks = [((wrapped_lon < lon_upper) & (wrapped_lon >= lon_lower)) \
                 for (lon_lower, lon_upper) in lon_edges]

    
    
    if return_pandas_dataframe:
        try:
            df = pd.DataFrame({
                "INTEN":data["INTEN"].byteswap().newbyteorder(), 
                "INTEN_DERED":data["INTEN_DERED"].byteswap().newbyteorder(),
                "tan(b)":data["tan(b)"].byteswap().newbyteorder(),
                "GAL-LON":data["GAL-LON"].byteswap().newbyteorder(),
                "GAL-LAT":data["GAL-LAT"].byteswap().newbyteorder(),
                "Av":data["Av"].byteswap().newbyteorder(),
                "DISTANCE":data["DISTANCE"],
                "Z":data["Z"]
            })
        except KeyError:
            df = pd.DataFrame({
                "INTEN":data["INTEN"].byteswap().newbyteorder(), 
                "tan(b)":data["tan(b)"].byteswap().newbyteorder(),
                "GAL-LON":data["GAL-LON"].byteswap().newbyteorder(),
                "GAL-LAT":data["GAL-LAT"].byteswap().newbyteorder(),
            })
        if longitude_mask_width is None:
            return data, df
        else:
            return data, df, masks
    else:
        return data


def fit_scale_heights(data, masks, min_lat = None, max_lat = None, 
    deredden = False, fig_names = None, return_smoothed = False, 
    smoothed_width = None, xlim = None, ylim = None):
    """
    Fits scale height data and returns slopes

    Parameters
    ----------
    data: `skySurvey`
        WHAM skySurvey object of full sky (requires track keyword), or spiral arm section
    masks: `list like`
        longitude masks to use
    min_lat:   `u.Quantity`
        min latitude to fit
    max_lat:   `u.Quantity`
        max latitude to fit
    deredden: `bool`
        if True, also fits dereddened slopes
    fig_names: `str`
        if provided, saves figures following this name
    return_smoothed: `bool`
        if True, returns smoothed longitude and slope estimates
    smoothed_width: `u.Quantity`
        width to smooth data to in longitude
    """

    # Default values
    if min_lat is None:
        min_lat = 5*u.deg
    elif not hasattr(min_lat, "unit"):
        min_lat *= u.deg

    if max_lat is None:
        max_lat = 35*u.deg
    elif not hasattr(max_lat, "unit"):
        max_lat *= u.deg   

    if smoothed_width is None:
        smoothed_width = 5*u.deg
    elif not hasattr(smoothed_width, "unit"):
        smoothed_width *= u.deg

    #initialize data arrays

    slopes_pos = []
    slopes_neg = []
    slopes_pos_dr = []
    slopes_neg_dr = []
    intercept_pos = []
    intercept_neg = []
    intercept_pos_dr = []
    intercept_neg_dr = []
    median_longitude = []
    median_distance = []
    for ell2 in range(len(masks)):
        xx = data["tan(b)"][masks[ell2]]
        yy = np.log(data["INTEN"][masks[ell2]])
        nan_mask = np.isnan(yy)
        
        if deredden:
            zz = np.log(data["INTEN_DERED"][masks[ell2]])
            nan_mask_z = np.isnan(zz)

        median_longitude.append(np.median(data["GAL-LON"][masks[ell2]]))
        if deredden:
            median_distance.append(np.median(data["DISTANCE"][masks[ell2]]))

        y_min = np.tan(min_lat)
        y_max = np.tan(max_lat)
        
        if hasattr(stats, "siegelslopes"):
            slope_estimator = stats.siegelslopes
        else:
            logging.warning("Installed version of scipy does not have the siegelslopes method in scipy.stats!")
            slope_estimator = stats.theilslopes

        siegel_result_pos = slope_estimator(yy[(xx > y_min) & (xx < y_max) & ~nan_mask],
                                               xx[(xx > y_min) & (xx < y_max) & ~nan_mask])
        siegel_result_neg = slope_estimator(yy[(xx < -y_min) & (xx > -y_max) & ~nan_mask],
                                               xx[(xx < -y_min) & (xx > -y_max) & ~nan_mask])
        
        if deredden:
            siegel_result_pos_dr = slope_estimator(zz[(xx > y_min) & (xx < y_max) & ~nan_mask_z],
                                                   xx[(xx > y_min) & (xx < y_max) & ~nan_mask_z])
            siegel_result_neg_dr = slope_estimator(zz[(xx < -y_min) & (xx > -y_max) & ~nan_mask_z],
                                                   xx[(xx < -y_min) & (xx > -y_max) & ~nan_mask_z])

        slopes_pos.append(siegel_result_pos[0])
        slopes_neg.append(siegel_result_neg[0])
        
        intercept_pos.append(siegel_result_pos[1])
        intercept_neg.append(siegel_result_neg[1])

        if deredden:
            slopes_pos_dr.append(siegel_result_pos_dr[0])
            slopes_neg_dr.append(siegel_result_neg_dr[0])
            intercept_pos_dr.append(siegel_result_pos_dr[1])
            intercept_neg_dr.append(siegel_result_neg_dr[1])

        if fig_names is not None:
            figure_name = "{0}_{1}.png".format(fig_names, ell2)

            if xlim is None:
                xlim = np.array([-0.9, 0.9])
            if ylim is None:
                ylim = np.array([-4.6, 3.2])

            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax2 = ax.twiny()


            ax.scatter(xx, 
                       yy, 
                       color ="k", 
                       alpha = 0.8)
            if deredden:
                ax.scatter(xx, 
                           zz, 
                           color ="grey", 
                           alpha = 0.8)


            ax.set_xlabel(r"$\tan$(b)", fontsize= 12)
            ax.set_ylabel(r"$\log$($H\alpha$ Intensity / R)", fontsize= 12)

            ax.set_title(r"${0:.1f} < l < {1:.1f}$".format(data["GAL-LON"][masks[ell2]].min(), 
                                                           data["GAL-LON"][masks[ell2]].max()), 
                         fontsize = 14)

            ax2.plot(np.degrees(np.arctan(xlim)), 
                     np.log([0.1,0.1]), ls = ":", lw = 1, 
                    color = "k", label = "0.1 R")
            ax2.fill_between([-min_lat, min_lat]*u.deg, [ylim[0], ylim[0]], [ylim[1], ylim[1]],
                            color = pal[1], 
                            alpha = 0.1, 
                            label = r"$|b| < 5\degree$")

            line_xx = np.linspace(y_min, y_max, 10)
            line_yy_pos = siegel_result_pos[0] * line_xx + siegel_result_pos[1]
            line_yy_neg = siegel_result_neg[0] * -line_xx + siegel_result_neg[1]
            ax.plot(line_xx, line_yy_pos, color = "r", lw = 3, alpha = 0.9, 
                    label = r"$H_{{n_e^2}} = {0:.2f} D$".format(1/-siegel_result_pos[0]))
            ax.plot(-line_xx, line_yy_neg, color = "b", lw = 3, alpha = 0.9, 
                    label = r"$H_{{n_e^2}} = {0:.2f} D$".format(1/siegel_result_neg[0]))

            if deredden:
                line_yy_pos_dr = siegel_result_pos_dr[0] * line_xx + siegel_result_pos_dr[1]
                line_yy_neg_dr = siegel_result_neg_dr[0] * -line_xx + siegel_result_neg_dr[1]
                ax.plot(line_xx, line_yy_pos_dr, color = "r", lw = 3, alpha = 0.9, ls = "--", 
                        label = r"Dered: $H_{{n_e^2}} = {0:.2f} D$".format(1/-siegel_result_pos_dr[0]))
                ax.plot(-line_xx, line_yy_neg_dr, color = "b", lw = 3, alpha = 0.9,  ls = "--", 
                        label = r"Dered: $H_{{n_e^2}} = {0:.2f} D$".format(1/siegel_result_neg_dr[0]))

            
            
            

            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

            ax2.set_xlabel(r"$b$ (deg)", fontsize = 12)
            ax2.set_xlim(np.degrees(np.arctan(xlim)))

            ax.legend(fontsize = 12, loc = 1)
            ax2.legend(fontsize = 12, loc = 2)

            plt.tight_layout()

            plt.savefig(figure_name, dpi = 300)
            del(fig)
            plt.close()

    results = {
    "median_longitude":np.array(median_longitude),
    "slopes_pos":np.array(slopes_pos),
    "slopes_neg":np.array(slopes_neg),
    "intercept_pos":np.array(intercept_pos),
    "intercept_neg":np.array(intercept_neg)
    }

    if deredden:

        results["median_distance"] = np.array(median_distance),
        results["slopes_pos_dr"] = np.array(slopes_pos_dr)
        results["slopes_neg_dr"] = np.array(slopes_neg_dr)
        results["intercept_pos_dr"] = np.array(intercept_pos_dr)
        results["intercept_neg_dr"] = np.array(intercept_neg_dr)

    if return_smoothed:
        results["smoothed_longitude"] = np.arange(np.min(median_longitude), 
            np.max(median_longitude), 0.25)
        if deredden:
            distance_interp = interp1d(median_longitude, median_distance)
            results["smoothed_distance"] = distance_interp(results["smoothed_longitude"])
        smoothed_slope_pos_ha = np.zeros((3,len(results["smoothed_longitude"])))
        smoothed_slope_neg_ha = np.zeros((3,len(results["smoothed_longitude"])))
        smoothed_slope_pos_ha_dr = np.zeros((3,len(results["smoothed_longitude"])))
        smoothed_slope_neg_ha_dr = np.zeros((3,len(results["smoothed_longitude"])))
        for ell,lon in enumerate(results["smoothed_longitude"]):
            smoothed_slope_pos_ha[:,ell] = np.nanpercentile(np.array(slopes_pos)[(median_longitude <= lon + smoothed_width.value/2) & 
                                                             (median_longitude > lon - smoothed_width.value/2)], 
                                               (10, 50, 90))
            smoothed_slope_neg_ha[:,ell] = np.nanpercentile(np.array(slopes_neg)[(median_longitude <= lon + smoothed_width.value/2) & 
                                                             (median_longitude > lon - smoothed_width.value/2)], 
                                               (10, 50, 90))
            if deredden:
                smoothed_slope_pos_ha_dr[:,ell] = np.nanpercentile(np.array(slopes_pos_dr)[(median_longitude <= lon + smoothed_width.value/2) & 
                                                                 (median_longitude > lon - smoothed_width.value/2)], 
                                                   (10, 50, 90))
                smoothed_slope_neg_ha_dr[:,ell] = np.nanpercentile(np.array(slopes_neg_dr)[(median_longitude <= lon + smoothed_width.value/2) & 
                                                                 (median_longitude > lon - smoothed_width.value/2)], 
                                                   (10, 50, 90))

        results["smoothed_slopes_pos"] = smoothed_slope_pos_ha
        results["smoothed_slopes_neg"] = smoothed_slope_neg_ha
        if deredden:
            results["smoothed_slopes_pos_dr"] = smoothed_slope_pos_ha_dr
            results["smoothed_slopes_neg_dr"] = smoothed_slope_neg_ha_dr

        

    return results
                              



import pytest
import numpy as np
from ..skySurvey import SkySurvey
from ..skySurvey import directory
# Set up the random number generator
np.random.seed(1234)

# Load global survey
survey = SkySurvey()

def test_get_spiral_slice():
    """
    basic spiral slice test
    """
    spiral_arm = survey.get_spiral_slice(track = "perseus")
    spiral_arm2 = survey.get_spiral_slice(track = "Per")

    assert np.allclose(spiral_arm["INTEN"], spiral_arm2["INTEN"], equal_nan = True)

def test_get_spiral_slice_by_file():
    """
    load custom data file with less than 6 columns
    """
    import os.path
    filename = os.path.join(directory, "tests", "test_data", "test_lbv.dat")
    spiral_arm = survey.get_spiral_slice(filename = filename)

    track = np.array([np.arange(10)+1, np.arange(10)+1, np.arange(10)+1]).T
    spiral_arm2 = survey.get_spiral_slice(track = track)

    assert np.allclose(spiral_arm["INTEN"], spiral_arm2["INTEN"], equal_nan = True)

def test_custom_lon_lat_range():
    """
    ensure custom bounds work
    """
    import astropy.units as u
    lrange = [-50,-40]
    brange = [-10,10]
    spiral_arm = survey.get_spiral_slice(track = "Carina_far", 
        lrange = lrange, 
        brange = brange)
    spiral_arm2 = survey.get_spiral_slice(track = "CrF", 
        lrange = lrange*u.deg, 
        brange = brange*u.deg)

    assert np.allclose(spiral_arm["INTEN"], spiral_arm2["INTEN"], equal_nan = True)

def test_track_shape_error():
    """
    ensure TypeError raised if bad track format provided
    """
    track = np.random.randn(50,1)
    try:
        bad_arm = survey.get_spiral_slice(track = track)
    except TypeError:
        assert True
    else:
        assert False

def test_track_lv_only():
    """
    ensure 2 column track shape works
    """
    track = np.arange(20).reshape(10,2)
    track2 = np.zeros((10,6))
    track2[:,0] = track[:,0]
    track2[:,2] = track[:,1]
    spiral_arm = survey.get_spiral_slice(track = track)
    spiral_arm2 = survey.get_spiral_slice(track = track2)

    assert np.allclose(spiral_arm["INTEN"], spiral_arm2["INTEN"], equal_nan = True)
    

def test_no_specification_error():
    """
    ensure SyntaxError raised if no track or filename provided
    """
    try:
        bad_arm = survey.get_spiral_slice()
    except SyntaxError:
        assert True
    else:
        assert False

def test_no_track_data_error():
    """
    ensure KeyError raised if bad track format provided
    """
    try:
        bad_arm = survey.get_spiral_slice(track = "fakeSpiral")
    except KeyError:
        assert True
    else:
        assert False

def test_vel_width():
    """
    ensure different velocity widths work
    """
    import astropy.units as u
    spiral_arm = survey.get_spiral_slice(track = "Perseus", 
        vel_width = 20.)
    spiral_arm2 = survey.get_spiral_slice(track = "Perseus", 
        vel_width = 20.*u.km/u.s)

    assert np.allclose(spiral_arm["INTEN"], spiral_arm2["INTEN"], equal_nan = True)

def test_wrap_at_180():
    """
    ensure longitude angle wrapping works
    """
    spiral_arm = survey.get_spiral_slice(track = "CrF", 
        wrap_at_180 = True)
    spiral_arm2 = survey.get_spiral_slice(track = "Carina_far", 
        wrap_at_180 = True)

    assert np.allclose(spiral_arm["INTEN"], spiral_arm2["INTEN"], equal_nan = True)

def test_interpolation():
    """
    ensure interpolation of central velocities work
    current a "fake" placeholder test - not yet implemented
    """
    spiral_arm = survey.get_spiral_slice(track = "perseus", 
        interpolate = True)
    spiral_arm2 = survey.get_spiral_slice(track = "Per", 
        interpolate = False)

    assert np.allclose(spiral_arm["INTEN"], spiral_arm2["INTEN"], equal_nan = True)

def test_return_track():
    """
    ensure track is returned if called
    """
    from ..lbvTracks import get_lbv_track
    spiral_arm, track = survey.get_spiral_slice(track = "carina_near", 
        return_track = True)
    track2 = get_lbv_track(reid_track = "carina_near")


    assert np.allclose(track, track2, equal_nan = True)







import pytest
import numpy as np
from ..skySurvey import SkySurvey
from ..skySurvey import directory
# Set up the random number generator
np.random.seed(1234)

# Load global survey
survey = SkySurvey()

def test_basic_input_track():
    """
    ensure inputting track without it pre-set as an attribute works
    """
    data, track = survey.get_spiral_slice(track = "Per", return_track = True)
    data2 = survey.get_scale_height_data(track = "Per", return_track = False)
    data.get_scale_height_data()

    assert np.allclose(data["INTEN"], data2["INTEN"], equal_nan = True)

def test_angle_wrapping():
    """
    ensure angle wrapping works properly
    """
    data = survey.get_scale_height_data(track = 'CrF', wrap_at_180 = True)
    data2 = survey.get_scale_height_data(track = 'CrF', wrap_at_180 = False)

    assert np.allclose(data["INTEN"], data2["INTEN"], equal_nan = True)

def test_no_track_error():
    """
    ensure SyntaxError when no track present / given
    """
    try:
        fail = survey.get_scale_height_data()
    except SyntaxError:
        assert True
    else:
        assert False

def test_deredden():
    """
    ensure different deredden keyword formats work
    """
    # from dustmaps.marshall import MarshallQuery
    # data = survey.get_scale_height_data(track = 'SgN', deredden = True)
    # data2 = survey.get_scale_height_data(track = "SgN", deredden = MarshallQuery())

    # assert np.allclose(data["INTEN"], data2["INTEN"], equal_nan = True)
    from dustmaps.bayestar import BayestarWebQuery
    try:
        data = survey.get_scale_height_data(track = 'SgN', deredden = True)
    except OSError:
        assert True
    else:
        data = survey.get_scale_height_data(track = 'SgN', deredden = BayestarWebQuery())
        assert True

def test_deredden_invalid():
    """
    ensure TypeError when invalid deredden keyword provided
    """
    try:
        data = survey.get_scale_height_data(track = "SgN", deredden = "break")
    except TypeError:
        assert True
    else:
        assert False

def test_bad_track():
    """
    ensure ValueError raised in track doesn't have enough columns
    """
    track = np.random.randn(10,2)
    try: 
        fail = survey.get_scale_height_data(track = track)
    except ValueError:
        assert True
    else:
        assert False

def test_pandas_dataframe():
    """
    ensure returning pandas dataframe works for both dereddened and raw data
    """
    from dustmaps.bayestar import BayestarWebQuery
    data, df = survey.get_scale_height_data(track = 'SgN', deredden = False, 
                                            return_pandas_dataframe = True)
    # data2, df2 = survey.get_scale_height_data(track = "SgN", deredden = False, 
    #                                           return_pandas_dataframe = True)

    # assert np.allclose(df["INTEN"], df2["INTEN"], equal_nan = True)
    data2, df2 = survey.get_scale_height_data(track = "SgN", 
            deredden = BayestarWebQuery(), 
            return_pandas_dataframe = True)
    assert np.allclose(df["INTEN"], df2["INTEN"], equal_nan = True)

def test_longitude_mask_width():
    """
    ensure returning masks works
    """
    import astropy.units as u
    data, df, masks = survey.get_scale_height_data(track = 'SgN',
                                            return_pandas_dataframe = True, 
                                            longitude_mask_width = 3)
    data2, df2, masks2 = survey.get_scale_height_data(track = "SgN", 
                                              return_pandas_dataframe = True,
                                              longitude_mask_width = 3*u.deg)

    assert np.allclose(df["INTEN"][masks[0]], df2["INTEN"][masks[0]], equal_nan = True)

def test_longitude_step_size():
    """
    ensure step size works
    """
    from dustmaps.bayestar import BayestarWebQuery
    import astropy.units as u
    data2, df2, masks2 = survey.get_scale_height_data(track = "SgN", deredden = False, 
                                              return_pandas_dataframe = True,
                                              longitude_mask_width = 5*u.deg, 
                                              step_size = 1)

    # assert np.allclose(df["INTEN"][masks[0]], df2["INTEN"][masks[0]], equal_nan = True) 
    
    data, df, masks = survey.get_scale_height_data(track = 'SgN', deredden = BayestarWebQuery(), 
                                        return_pandas_dataframe = True, 
                                        longitude_mask_width = 5, 
                                        step_size = 1*u.deg)
    assert np.allclose(df["INTEN"][masks[0]], df2["INTEN"][masks[0]], equal_nan = True)

def test_add_kinematic():
    """
    ensure adding kinematic distance works
    """
    data, df = survey.get_scale_height_data(track = "CrN", 
        return_pandas_dataframe = True, add_kinematic_distance = True)
    data2, df2 = survey.get_scale_height_data(track = "CrN", 
        return_pandas_dataframe = True, add_kinematic_distance = True, 
        closer = True)

    assert np.allclose(df["INTEN"], df2["INTEN"], equal_nan = True)

def test_deredden_web():
    """
    Ensure deredden works using BayestarWebQuery
    """
    from dustmaps.bayestar import BayestarWebQuery
    data = survey.get_scale_height_data(track = 'SgN', deredden = BayestarWebQuery())
    assert np.any(~np.isnan(data["INTEN_DERED"]))






import logging

import numpy as np 
from astropy import units as u
from astropy.table import Table
from astropy.io import fits
from astropy.coordinates import SkyCoord, Angle

from .whampyTableMixin import SkySurveyMixin
from scipy.io import readsav

import os.path 

directory = os.path.dirname(__file__)



class SkySurvey(SkySurveyMixin, Table):
    """
    Core WHAM SkySurvey CLass

    Load, view, manipulate, and plot basic results from WHAM Sky Survey

    Parameters
    ----------
    filename: 'str', optional, must be keyword
        filename of WHAM survey fits file
        Defaults to URL
        can also be idl sav file
    mode: 'str', optional, must be keyword
        can be "local" or "remote" to signify using local file or remote access
    idl_var: 'str', optional, must be keyword
        if provided, points to name of IDL structure in sav file containing relevant data
        by default, will take the first variable that is a `numpy.recarray`
    file_list: 'str', 'listlike', optional, must be keyword
        if provided, reads combo fits files directly
    extension: 'str', optional, must be keyword
        extension to load from fits file
        deftaul to "ATMSUB"
    from_table: `astropy.table.Table`, `dictionary`, optional, must be keyword
        if provided, initializes directly from a Table
    tuner_list: 'str', 'listlike', optional, must be keyword
        if provided, reads tuner spectra fits files directly


    """


    def __init__(self, filename = None, mode = 'local', idl_var = None, 
                 file_list = None, extension = None, 
                 from_table = None, tuner_list = None,
                 **kwargs):


        if from_table is not None:
            super().__init__(data = from_table, **kwargs)

        elif tuner_list is not None:
            if isinstance(tuner_list, str):
                tuner_list = [tuner_list]
            elif not hasattr(tuner_list, "__iter__"):
                raise TypeError("Invalid tuner_list Type, tuner_list must be a string, list of strings, or None.")

            if not isinstance(tuner_list, np.ndarray):
                tuner_list = np.array(tuner_list)

            if extension is None:
                extension = "RAWSPEC"

            file_dict = {}
            for ell, filename in enumerate(tuner_list):
                with fits.open(filename) as hdulist:
                    primary_header = hdulist["PRIMARY"].header
                    extension_header = hdulist[extension].header
                    extension_data = hdulist[extension].data

                # Read in Spectra
                for key in extension_data.dtype.names:
                    if ell > 0:
                        file_dict[key].append(extension_data[key])
                    else:
                        file_dict[key] = [extension_data[key]]

                # Read in some Primary Header Data
                for key in primary_header.keys():
                    if key not in extension_header.keys():
                        if ell > 0:
                            file_dict[key].append(primary_header[key])
                        else:
                            file_dict[key] = [primary_header[key]]

            # Units for Data
            file_dict["VELOCITY"] *= u.km/u.s
            file_dict["DATA"] *= u.R / 22.8 # ADU to R
            file_dict["VARIANCE"] *= u.R**2 / 22.8**2

            super().__init__(data = file_dict, **kwargs)

            if "INTEN" not in file_dict:
                self["INTEN"] = self.moment()
            if "ERROR" not in file_dict:
                _, self["ERROR"] = self.moment(return_sigma = True)

            del file_dict
                       
        

        # Check calibrators keyword
        elif file_list is not None:
            if isinstance(file_list, str):
                file_list = [file_list]
            elif not hasattr(file_list, "__iter__"):
                raise TypeError("Invalid file_list Type, file_list must be a string, list of strings, or None.")

            if not isinstance(file_list, np.ndarray):
                file_list = np.array(file_list)

            if extension is None:
                extension = "ATMSUB"

            # Check extension exists
            extension_exists_mask = []
            for filename in file_list:
                with fits.open(filename) as hdulist:
                    extension_exists_mask.append(extension in hdulist)

            if np.sum(extension_exists_mask) < len(file_list):
                logging.warning("Not all files in file_list have the extension: {}!".format(extension))

            if np.sum(extension_exists_mask) == 0:
                raise ValueError("No files in file_list have the extension: {}!".format(extension))

            file_dict = {}
            for ell, filename in enumerate(file_list[extension_exists_mask]):

                with fits.open(filename) as hdulist:
                    primary_header = hdulist["PRIMARY"].header
                    atmsub_header = hdulist[extension].header
                    atmsub_data = hdulist[extension].data

                # Read in ATMSUB Header Data
                for key in atmsub_header.keys():
                    if ell > 0:
                        file_dict[key].append(atmsub_header[key])
                    else:
                        file_dict[key] = [atmsub_header[key]]

                # Read in Spectra
                for key in atmsub_data.dtype.names:
                    if ell > 0:
                        file_dict[key].append(atmsub_data[key])
                    else:
                        file_dict[key] = [atmsub_data[key]]

                # Read in some Primary Header Data
                for key in primary_header.keys():
                    if key not in atmsub_header.keys():
                        if ell > 0:
                            file_dict[key].append(primary_header[key])
                        else:
                            file_dict[key] = [primary_header[key]]



            # Units for Data
            file_dict["VELOCITY"] *= u.km/u.s
            file_dict["DATA"] *= u.R / 22.8 # ADU to R
            file_dict["VARIANCE"] *= u.R**2 / 22.8**2

            super().__init__(data = file_dict, **kwargs)

            if "INTEN" not in file_dict:
                self["INTEN"] = self.moment()
            if "ERROR" not in file_dict:
                _, self["ERROR"] = self.moment(return_sigma = True)

            del file_dict
            


        else:


            if filename is None:
                if mode == 'local':
                    filename = os.path.join(directory, "data/wham-ss-DR1-v161116-170912.fits")
                elif mode == 'remote':
                    filename = "http://www.astro.wisc.edu/wham/ss/wham-ss-DR1-v161116-170912.fits"

            if filename[-4:] == ".sav":
                # IDL Save File
                idl_data = readsav(filename)

                # Find right data entry
                if idl_var is None:
                    for key in idl_data.keys():
                        if idl_data[key].__class__ is np.recarray:
                            idl_var = key
                            break
                if idl_var not in idl_data.keys():
                    raise TypeError("Could not find WHAM data structure in IDL Save File")

                survey_data = idl_data[idl_var]

                # Covert some columns to float arrays and add to new dictionary
                data_dict = {}
                # Standard WHAM IDL Save File Format
                data_dict["GAL-LON"] = survey_data["GLON"] * u.deg
                data_dict["GAL-LAT"] = survey_data["GLAT"] * u.deg
                data_dict["VELOCITY"] = np.vstack(survey_data["VEL"][:][:]) * u.km/u.s
                data_dict["DATA"] = np.vstack(survey_data["DATA"][:][:]) * u.R * u.s / u.km
                data_dict["VARIANCE"] = np.vstack(survey_data["VAR"][:][:]) * (u.R * u.s / u.km)**2

                # Extra info from IDL Save Files Dependent on type of File / origin
                if "INTEN" in survey_data.dtype.names:
                    data_dict["INTEN"] = survey_data["INTEN"] * u.R
                if "OINTEN" in survey_data.dtype.names:
                    data_dict["OINTEN"] = survey_data["OINTEN"] * u.R
                if "ERROR" in survey_data.dtype.names:
                    data_dict["ERROR"] = survey_data["ERROR"] * u.R

                for name in survey_data.dtype.names:
                    if name not in ("GLON", "GLAT", "VEL", "DATA", "VAR", "INTEN", "OINTEN", "ERROR"):
                        data_dict[name] = survey_data[name]

                super().__init__(data = data_dict, **kwargs)

                if "INTEN" not in data_dict:
                    self["INTEN"] = self.moment()
                if "ERROR" not in data_dict:
                    _, self["ERROR"] = self.moment(return_sigma = True)

                del data_dict
                del idl_data



            else:
                with fits.open(filename) as hdulist:
                    self.header = hdulist[0].header
                    self.table_header = hdulist[1].header

                    self.date = hdulist[0].header["DATE"]
                    if "TAG" in hdulist[0].header.keys():
                        self.tag = hdulist[0].header["TAG"]
                    if "VERSION" in hdulist[0].header.keys():
                        self.version = hdulist[0].header["VERSION"]

                t = Table.read(filename)

                # Set / Fix Units to comply with astropy.units
                for column in t.columns:
                    if t[column].unit == "DEG":
                        t[column].unit = u.deg
                    elif t[column].unit == "KM/S":
                        t[column].unit = u.km/u.s 
                    elif t[column].unit == "RAYLEIGH/(KM/S)":
                        t[column].unit = u.R / u.km * u.s
                    elif t[column].unit == "(RAYLEIGH/(KM/S))^2":
                        t[column].unit = (u.R / u.km * u.s)**2
                    elif t[column].unit == "RAYLEIGH":
                        t[column].unit = u.R

                super().__init__(data = t.columns, meta = t.meta, **kwargs)

                del t



    def sky_section(self, bounds, radius = None, wrap_at_180 = True):
        """
        Extract a sub section of the survey from the sky

        Parameters
        ----------

        bounds: `list` or `Quantity` or `SkyCoord`
            if `list` or `Quantity` must be formatted as:
                [min Galactic Longitude, max Galactic Longitude, min Galactic Latitude, max Galactic Latitude]
                or 
                [center Galactic Longitude, center Galactic Latitude] and requires radius keyword to be set
                default units of u.deg are assumed
            if `SkyCoord', must be length 4 or length 1 or length 2
                length 4 specifies 4 corners of rectangular shape
                length 1 specifies center of circular region and requires radius keyword to be set
                length 2 specifies two corners of rectangular region
        radius: 'number' or 'Quantity', optional, must be keyword
            sets radius of circular region
        wrap_at_180: `bool`, optional, must be keyword
            if True, wraps longitude angles at 180d
            use if mapping accross Galactic Center
        """
        if wrap_at_180:
            wrap_at = "180d"
        else:
            wrap_at = "360d"

        if not isinstance(bounds, u.Quantity) | isinstance(bounds, SkyCoord):
            bounds *= u.deg
            logging.warning("No units provided for bounds, assuming u.deg")

        wham_coords = self.get_SkyCoord()

        if isinstance(bounds, SkyCoord):
            if len(bounds) == 1:
                if radius is None:
                    raise TypeError("Radius must be provided if only a single coordinate is given")
                elif not isinstance(radius, u.Quantity):
                    radius *= u.deg
                    logging.warning("No units provided for radius, assuming u.deg")
                center = bounds
            elif len(bounds) >= 2:
                min_lon, max_lon = bounds.l.wrap_at(wrap_at).min(), bounds.l.wrap_at(wrap_at).max()
                min_lat, max_lat = bounds.b.min(), bounds.l.max()
        elif len(bounds) == 2:
            if radius is None:
                raise TypeError("Radius must be provided if only a single coordinate is given")
            elif not isinstance(radius, u.Quantity):
                radius *= u.deg
                logging.warning("No units provided for radius, assuming u.deg")
            center = SkyCoord(l = bounds[0], b = bounds[1], frame = 'galactic')
        elif len(bounds) == 4:
            min_lon, max_lon, min_lat, max_lat = Angle(bounds)
            min_lon = min_lon.wrap_at(wrap_at)
            max_lon = max_lon.wrap_at(wrap_at)
        else:
            raise TypeError("Input bounds and/or radius are not understood")

        # rectangular extraction
        if radius is None:
            # Mask of points inside rectangular region
            inside_mask = wham_coords.l.wrap_at(wrap_at) <= max_lon
            inside_mask &= wham_coords.l.wrap_at(wrap_at) >= min_lon
            inside_mask &= wham_coords.b <= max_lat
            inside_mask &= wham_coords.b >= min_lat

        else: # Circle extraction
            # Compute Separation
            # Warning to self: This is VERY slow
            sep = wham_coords.separation(center)

            # Mask of points inside circular region
            inside_mask = sep <= radius

        return self[inside_mask]



from astropy import units as u
import numpy as np 
import matplotlib.pyplot as plt


try:
    from spectral_cube import SpectralCube
except ModuleNotFoundError:
    # Error handling
    pass
try:
    import cartopy.crs as ccrs
except ModuleNotFoundError:
    # Error handling
    pass

from astropy.coordinates import SkyCoord
from astropy.coordinates import Angle





class SpectrumPlotter():
    """
    Class to interactivly plot spectra from WHAM maps
    Click on a point on a map and a spectrum will be extracted cooresponding to the 
    closest point on the Sky

    Parameters
    ----------

    image_ax: 'axes` 
        containing the image / map'
    line:   'matplotlib.line.Line2D'
        line that will be updated to match clicked spectra
    data:   'SkySurvey', optional, must be keyword
        WHAM data
    over_data:  'SkySurvey' or 'str' or 'spectral_cube.SpectralCube', optional, must be keyword
        Extra data to over plot spectra from
        if SkySurvey, assumed to be WHAM observations, perhaps at another wavelength
        if 'str', assumed to be a 3D FITS Data cube filename to be loaded as SpectralCube
        if 'SpectralCube', defaults to extracting closest spectra to click
    over_line: matplotlib.line.Line2D', optional, must be keyword
        if over_data, must be provided
        line that will be udpated to match clicked spectra from over_data
    average_beam: 'bool', optional, must be keyword
        if True, instead over plots average spectrum from over_data within nearest WHAM beam
    radius: 'Quantity' or  'number', optional, must be keyword
        beam radius to average beam over if avreage_beam is True
        default is 0.5 degrees (WHAM beam)





    """
    def __init__(self, image_ax, line, data = None, over_data = None, over_line = None, 
        average_beam = False, radius = None):
        self.line = line
        self.image_ax = image_ax
        self.line_ax = line.axes
        self.over_data = over_data
        self.average_beam = average_beam
        self.radius = radius

        if hasattr(self.image_ax, 'coastlines'):
            self.geo = True
        else:
            self.geo = False
        if hasattr(self.image_ax, 'wcs'):
            self.wcs_axes = True
        else:
            self.wcs_axes = False
        self.data = data
        if self.over_data is not None:
            if over_data.__class__ is str:
                self.over_data = SpectralCube.read(over_data)
                self.over_type = "SpectralCube"
            elif hasattr(over_data, 'minimal_subcube'):
                # Checks if it is a SpectralCube or wrapper class of SpectralCube
                self.over_data = over_data
                self.over_type = "SpectralCube"
            elif hasattr(over_data, 'intensity_map'):
                # Checks if it is a SkySurvey object
                self.over_data = over_data
                self.over_type = "SkySurvey"
            if self.average_beam:
                if self.radius is None:
                    self.radius = 0.5 * u.deg
                elif not isinstance(radius, u.Quantity):
                    self.radius = radius * u.deg # Assume Default Units
                else:
                    self.radius = radius
        self.over_line = over_line
        self.cid = self.line.figure.canvas.mpl_connect("button_press_event", self.on_click)
    def on_click(self, event):
        if event.button == 1: # left mouse click
            if event.inaxes is not self.image_ax: 
                return # if click is not on image, nothing happens
            else:
                lon = event.xdata
                lat = event.ydata
                if self.geo: # if cartopy axes with projection info
                    # Convert coordinates to standard longitude and latitude
                    lon, lat = ccrs.PlateCarree().transform_point(lon, lat, self.image_ax.projection)
                elif self.wcs_axes:
                    # Covert coordinates from pixel to world
                    lon, lat, _ = self.image_ax.wcs.wcs_pix2world(lon, lat, 0, 0)

                # Create SKyCoord
                click_coord = SkyCoord(l = lon*u.deg, b = lat*u.deg, frame = 'galactic')
                # Find closest Spectrum index
                closest = self.data.get_spectrum(click_coord, index = True)
                galCoord = self.data.get_SkyCoord()[closest]

                # Update line to be the spectral data
                self.line.set_data(self.data[closest]["VELOCITY"], 
                                   self.data[closest]["DATA"])
                if self.over_line is not None:
                    if self.over_line.axes == self.line_ax:
                        self.over_line.set_data([0],[0])
                self.line_ax.relim()
                self.line_ax.autoscale_view() # Rescale axes to cover data

                # Draw over_line if needed:
                if self.over_line is not None:
                    if self.over_type == 'SkySurvey':
                        closest = self.over_data.get_spectrum(click_coord, index = True)
                        galCoord = self.data.get_SkyCoord()[closest]

                        # Update line to be the spectral data
                        self.over_line.set_data(self.over_data[closest]["VELOCITY"], 
                                           self.over_data[closest]["DATA"])
                        self.over_line.axes.relim()
                        self.over_line.axes.autoscale_view() # Rescale axes to cover data

                    elif self.over_type == 'SpectralCube':
                        if not self.average_beam:
                            # find index of closest value
                            _, lat_axis_values, _ = self.over_data.world[int(self.over_data.shape[0]/2), :, 
                                                                            int(self.over_data.shape[2]/2)]
                            lat_slice = np.nanargmin(np.abs(lat_axis_values-click_coord.b))

                            _, _, lon_axis_values = self.over_data.world[int(self.over_data.shape[0]/2), 
                                                                            int(self.over_data.shape[1]/2), :]
                            # Ensure all angles are wrapped at 180
                            lon_axis_values = Angle(lon_axis_values).wrap_at("180d")
                            lon_slice = np.nanargmin(np.abs(lon_axis_values-click_coord.l.wrap_at("180d")))
                            self.over_line.set_data(self.over_data.spectral_axis.to(u.km/u.s).value, 
                                                    self.over_data.unmasked_data[:,lat_slice,lon_slice].value)
                        else:
                            ds9_str = 'galactic; circle({0:.3}, {1:.4}, {2:.4}")'.format(click_coord.l.wrap_at("180d").value, 
                                                                         click_coord.b.value, 
                                                                         self.radius.to(u.arcsec).value)
                            _, lat_axis_values, _ = self.over_data.world[int(self.over_data.shape[0]/2), :, int(self.over_data.shape[2]/2)]
                            lat_slice_up = np.nanargmin(np.abs(lat_axis_values-click_coord.b+self.radius*1.5))
                            lat_slice_down = np.nanargmin(np.abs(lat_axis_values-click_coord.b-self.radius*1.5))
                            lat_slices = np.sort([lat_slice_up, lat_slice_down])


                            _, _, lon_axis_values = self.over_data.world[int(self.over_data.shape[0]/2), int(self.over_data.shape[1]/2), :]
                            # Ensure all angles are wrapped at 180
                            lon_axis_values = Angle(lon_axis_values).wrap_at("180d")
                            lon_slice_up = np.nanargmin(np.abs(lon_axis_values-click_coord.l.wrap_at("180d")+self.radius*1.5))
                            lon_slice_down = np.nanargmin(np.abs(lon_axis_values-click_coord.l.wrap_at("180d")-self.radius*1.5))
                            lon_slices = np.sort([lon_slice_up, lon_slice_down])


                            smaller_cube = self.over_data[:,lat_slices[0]:lat_slices[1], lon_slices[0]:lon_slices[1]]
                            subcube = smaller_cube.subcube_from_ds9region(ds9_str)
                            spectrum = subcube.mean(axis = (1,2))
                            self.over_line.set_data(spectrum.spectral_axis.to(u.km/u.s).value, 
                                                    spectrum.value)
                        if self.over_line.axes != self.line_ax:
                            self.over_line.axes.relim()
                            self.over_line.axes.autoscale_view() # Rescale axes to cover data
                            if self.over_type == 'SkySurvey':
                                self.over_line.axes.set_ylabel("Intensity ({0})".format(self.over_data["DATA"].unit), 
                                                fontsize = 12)
                            else:
                                self.over_line.axes.set_ylabel("Intensity ({0})".format(self.over_data.unmasked_data[0,0,0].unit), 
                                                fontsize = 12)

                # Set Labels
                self.line_ax.set_ylabel("Intensity ({0})".format(self.data["DATA"].unit), 
                                        fontsize = 12)
                self.line_ax.set_xlabel("LSR Velocity ({0}) | (l,b) = ({1:3.1f},{2:3.1f})".format(self.data["VELOCITY"].unit, 
                                                                                                galCoord.l.wrap_at("180d").value, 
                                                                                                galCoord.b.value), 
                                        fontsize = 12)

                # Draw the Line
                self.line.figure.canvas.draw()
                


# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This directive will produce an "autosummary"-style table for public
attributes of a specified module. See the `sphinx.ext.autosummary`_ extension
for details on this process. The main difference from the `autosummary`_
directive is that `autosummary`_ requires manually inputting all attributes
that appear in the table, while this captures the entries automatically.

This directive requires a single argument that must be a module or
package.

It also accepts any options supported by the `autosummary`_ directive-
see `sphinx.ext.autosummary`_ for details. It also accepts some additional
options:

    * ``:classes-only:``
        If present, the autosummary table will only contain entries for
        classes. This cannot be used at the same time with
        ``:functions-only:`` or ``:variables-only:``.

    * ``:functions-only:``
        If present, the autosummary table will only contain entries for
        functions. This cannot be used at the same time with
        ``:classes-only:`` or ``:variables-only:``.

    * ``:variables-only:``
        If present, the autosummary table will only contain entries for
        variables (everything except functions and classes). This cannot
        be used at the same time with ``:classes-only:`` or
        ``:functions-only:``.

    * ``:skip: obj1, [obj2, obj3, ...]``
        If present, specifies that the listed objects should be skipped
        and not have their documentation generated, nor be included in
        the summary table.

    * ``:allowed-package-names: pkgormod1, [pkgormod2, pkgormod3, ...]``
        Specifies the packages that functions/classes documented here are
        allowed to be from, as comma-separated list of package names. If not
        given, only objects that are actually in a subpackage of the package
        currently being documented are included.

    * ``:inherited-members:`` or ``:no-inherited-members:``
        The global sphinx configuration option ``automodsumm_inherited_members``
        decides if members that a class inherits from a base class are included
        in the generated documentation. The flags ``:inherited-members:`` or
        ``:no-inherited-members:`` allows overrriding this global setting.

This extension also adds two sphinx configuration options:

* ``automodsumm_writereprocessed``
    Should be a bool, and if ``True``, will cause `automodsumm`_ to write files
    with any ``automodsumm`` sections replaced with the content Sphinx
    processes after ``automodsumm`` has run.  The output files are not
    actually used by sphinx, so this option is only for figuring out the
    cause of sphinx warnings or other debugging.  Defaults to ``False``.

* ``automodsumm_inherited_members``
    Should be a bool and if ``True``, will cause `automodsumm`_ to document
    class members that are inherited from a base class. This value can be
    overriden for any particular automodsumm directive by including the
    ``:inherited-members:`` or ``:no-inherited-members:`` options.  Defaults to
    ``False``.

.. _sphinx.ext.autosummary: http://sphinx-doc.org/latest/ext/autosummary.html
.. _autosummary: http://sphinx-doc.org/latest/ext/autosummary.html#directive-autosummary

.. _automod-diagram:

automod-diagram directive
=========================

This directive will produce an inheritance diagram like that of the
`sphinx.ext.inheritance_diagram`_ extension.

This directive requires a single argument that must be a module or
package. It accepts no options.

.. note::
    Like 'inheritance-diagram', 'automod-diagram' requires
    `graphviz <http://www.graphviz.org/>`_ to generate the inheritance diagram.

.. _sphinx.ext.inheritance_diagram: http://sphinx-doc.org/latest/ext/inheritance.html
"""

import inspect
import os
import re
import io
from distutils.version import LooseVersion

from sphinx import __version__
from sphinx.ext.autosummary import Autosummary
from sphinx.ext.inheritance_diagram import InheritanceDiagram
from docutils.parsers.rst.directives import flag

from .utils import find_mod_objs, cleanup_whitespace

SPHINX_LT_17 = LooseVersion(__version__) < LooseVersion('1.7')


def _str_list_converter(argument):
    """
    A directive option conversion function that converts the option into a list
    of strings. Used for 'skip' option.
    """
    if argument is None:
        return []
    else:
        return [s.strip() for s in argument.split(',')]


class Automodsumm(Autosummary):
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    has_content = False
    option_spec = dict(Autosummary.option_spec)
    option_spec['functions-only'] = flag
    option_spec['classes-only'] = flag
    option_spec['variables-only'] = flag
    option_spec['skip'] = _str_list_converter
    option_spec['allowed-package-names'] = _str_list_converter
    option_spec['inherited-members'] = flag
    option_spec['no-inherited-members'] = flag

    def run(self):
        env = self.state.document.settings.env
        modname = self.arguments[0]

        self.warnings = []
        nodelist = []

        try:
            localnames, fqns, objs = find_mod_objs(modname)
        except ImportError:
            self.warnings = []
            self.warn("Couldn't import module " + modname)
            return self.warnings

        try:
            # set self.content to trick the autosummary internals.
            # Be sure to respect functions-only and classes-only.
            funconly = 'functions-only' in self.options
            clsonly = 'classes-only' in self.options
            varonly = 'variables-only' in self.options
            if [clsonly, funconly, varonly].count(True) > 1:
                self.warning('more than one of functions-only, classes-only, '
                             'or variables-only defined. Ignoring.')
                clsonly = funconly = varonly = False

            skipnames = []
            if 'skip' in self.options:
                option_skipnames = set(self.options['skip'])
                for lnm in localnames:
                    if lnm in option_skipnames:
                        option_skipnames.remove(lnm)
                        skipnames.append(lnm)
                if len(option_skipnames) > 0:
                    self.warn('Tried to skip objects {objs} in module {mod}, '
                              'but they were not present.  Ignoring.'
                              .format(objs=option_skipnames, mod=modname))

            if funconly:
                cont = []
                for nm, obj in zip(localnames, objs):
                    if nm not in skipnames and inspect.isroutine(obj):
                        cont.append(nm)
            elif clsonly:
                cont = []
                for nm, obj in zip(localnames, objs):
                    if nm not in skipnames and inspect.isclass(obj):
                        cont.append(nm)
            elif varonly:
                cont = []
                for nm, obj in zip(localnames, objs):
                    if nm not in skipnames and not (inspect.isclass(obj) or
                                                    inspect.isroutine(obj)):
                        cont.append(nm)
            else:
                cont = [nm for nm in localnames if nm not in skipnames]

            self.content = cont

            # for some reason, even though ``currentmodule`` is substituted in,
            # sphinx doesn't necessarily recognize this fact.  So we just force
            # it internally, and that seems to fix things
            env.temp_data['py:module'] = modname
            env.ref_context['py:module'] = modname

            # can't use super because Sphinx/docutils has trouble return
            # super(Autosummary,self).run()
            nodelist.extend(Autosummary.run(self))

            return self.warnings + nodelist
        finally:  # has_content = False for the Automodsumm
            self.content = []

    def get_items(self, names):
        self.genopt['imported-members'] = True
        return Autosummary.get_items(self, names)


# <-------------------automod-diagram stuff----------------------------------->
class Automoddiagram(InheritanceDiagram):

    option_spec = dict(InheritanceDiagram.option_spec)
    option_spec['allowed-package-names'] = _str_list_converter
    option_spec['skip'] = _str_list_converter

    def run(self):
        try:
            ols = self.options.get('allowed-package-names', [])
            ols = True if len(ols) == 0 else ols  # if none are given, assume only local

            nms, objs = find_mod_objs(self.arguments[0], onlylocals=ols)[1:]
        except ImportError:
            self.warnings = []
            self.warn("Couldn't import module " + self.arguments[0])
            return self.warnings

        # Check if some classes should be skipped
        skip = self.options.get('skip', [])

        clsnms = []
        for n, o in zip(nms, objs):

            if n.split('.')[-1] in skip:
                continue

            if inspect.isclass(o):
                clsnms.append(n)

        oldargs = self.arguments
        try:
            if len(clsnms) > 0:
                self.arguments = [' '.join(clsnms)]
            return InheritanceDiagram.run(self)
        finally:
            self.arguments = oldargs


# <---------------------automodsumm generation stuff-------------------------->
def process_automodsumm_generation(app):
    env = app.builder.env

    filestosearch = []
    for docname in env.found_docs:
        filename = env.doc2path(docname)
        if os.path.isfile(filename):
            filestosearch.append(docname + os.path.splitext(filename)[1])

    liness = []
    for sfn in filestosearch:
        lines = automodsumm_to_autosummary_lines(sfn, app)
        liness.append(lines)
        if app.config.automodsumm_writereprocessed:
            if lines:  # empty list means no automodsumm entry is in the file
                outfn = os.path.join(app.srcdir, sfn) + '.automodsumm'
                with open(outfn, 'w') as f:
                    for l in lines:
                        f.write(l)
                        f.write('\n')

    for sfn, lines in zip(filestosearch, liness):
        suffix = os.path.splitext(sfn)[1]
        if len(lines) > 0:
            generate_automodsumm_docs(
                lines, sfn, app=app, builder=app.builder, warn=app.warn, info=app.info,
                suffix=suffix, base_path=app.srcdir,
                inherited_members=app.config.automodsumm_inherited_members)

# _automodsummrex = re.compile(r'^(\s*)\.\. automodsumm::\s*([A-Za-z0-9_.]+)\s*'
#                              r'\n\1(\s*)(\S|$)', re.MULTILINE)
_lineendrex = r'(?:\n|$)'
_hdrex = r'^\n?(\s*)\.\. automodsumm::\s*(\S+)\s*' + _lineendrex
_oprex1 = r'(?:\1(\s+)\S.*' + _lineendrex + ')'
_oprex2 = r'(?:\1\4\S.*' + _lineendrex + ')'
_automodsummrex = re.compile(_hdrex + '(' + _oprex1 + '?' + _oprex2 + '*)',
                             re.MULTILINE)


def automodsumm_to_autosummary_lines(fn, app):
    """
    Generates lines from a file with an "automodsumm" entry suitable for
    feeding into "autosummary".

    Searches the provided file for `automodsumm` directives and returns
    a list of lines specifying the `autosummary` commands for the modules
    requested. This does *not* return the whole file contents - just an
    autosummary section in place of any :automodsumm: entries. Note that
    any options given for `automodsumm` are also included in the
    generated `autosummary` section.

    Parameters
    ----------
    fn : str
        The name of the file to search for `automodsumm` entries.
    app : sphinx.application.Application
        The sphinx Application object

    Returns
    -------
    lines : list of str
        Lines for all `automodsumm` entries with the entries replaced by
        `autosummary` and the module's members added.


    """

    fullfn = os.path.join(app.builder.env.srcdir, fn)

    with io.open(fullfn, encoding='utf8') as fr:
        # Note: we use __name__ here instead of just writing the module name in
        #       case this extension is bundled into another package
        from . import automodapi

        try:
            extensions = app.extensions
        except AttributeError:  # Sphinx <1.6
            extensions = app._extensions

        if automodapi.__name__ in extensions:
            # Must do the automodapi on the source to get the automodsumm
            # that might be in there
            docname = os.path.splitext(fn)[0]
            filestr = automodapi.automodapi_replace(fr.read(), app, True, docname, False)
        else:
            filestr = fr.read()

    spl = _automodsummrex.split(filestr)
    # 0th entry is the stuff before the first automodsumm line
    indent1s = spl[1::5]
    mods = spl[2::5]
    opssecs = spl[3::5]
    indent2s = spl[4::5]
    remainders = spl[5::5]

    # only grab automodsumm sections and convert them to autosummary with the
    # entries for all the public objects
    newlines = []

    # loop over all automodsumms in this document
    for i, (i1, i2, modnm, ops, rem) in enumerate(zip(indent1s, indent2s, mods,
                                                      opssecs, remainders)):
        allindent = i1 + ('    ' if i2 is None else i2)

        # filter out functions-only, classes-only, and ariables-only
        # options if present.
        oplines = ops.split('\n')
        toskip = []
        allowedpkgnms = []
        funcsonly = clssonly = varsonly = False
        for i, ln in reversed(list(enumerate(oplines))):
            if ':functions-only:' in ln:
                funcsonly = True
                del oplines[i]
            if ':classes-only:' in ln:
                clssonly = True
                del oplines[i]
            if ':variables-only:' in ln:
                varsonly = True
                del oplines[i]
            if ':skip:' in ln:
                toskip.extend(_str_list_converter(ln.replace(':skip:', '')))
                del oplines[i]
            if ':allowed-package-names:' in ln:
                allowedpkgnms.extend(_str_list_converter(ln.replace(':allowed-package-names:', '')))
                del oplines[i]
        if [funcsonly, clssonly, varsonly].count(True) > 1:
            msg = ('Defined more than one of functions-only, classes-only, '
                   'and variables-only.  Skipping this directive.')
            lnnum = sum([spl[j].count('\n') for j in range(i * 5 + 1)])
            app.warn('[automodsumm]' + msg, (fn, lnnum))
            continue

        # Use the currentmodule directive so we can just put the local names
        # in the autosummary table.  Note that this doesn't always seem to
        # actually "take" in Sphinx's eyes, so in `Automodsumm.run`, we have to
        # force it internally, as well.
        newlines.extend([i1 + '.. currentmodule:: ' + modnm,
                         '',
                         '.. autosummary::'])
        newlines.extend(oplines)

        ols = True if len(allowedpkgnms) == 0 else allowedpkgnms
        for nm, fqn, obj in zip(*find_mod_objs(modnm, onlylocals=ols)):
            if nm in toskip:
                continue
            if funcsonly and not inspect.isroutine(obj):
                continue
            if clssonly and not inspect.isclass(obj):
                continue
            if varsonly and (inspect.isclass(obj) or inspect.isroutine(obj)):
                continue
            newlines.append(allindent + nm)

    # add one newline at the end of the autosummary block
    newlines.append('')

    return newlines


def generate_automodsumm_docs(lines, srcfn, app=None, suffix='.rst', warn=None,
                              info=None, base_path=None, builder=None,
                              template_dir=None,
                              inherited_members=False):
    """
    This function is adapted from
    `sphinx.ext.autosummary.generate.generate_autosummmary_docs` to
    generate source for the automodsumm directives that should be
    autosummarized. Unlike generate_autosummary_docs, this function is
    called one file at a time.
    """

    from sphinx.jinja2glue import BuiltinTemplateLoader
    from sphinx.ext.autosummary import import_by_name, get_documenter
    from sphinx.ext.autosummary.generate import (_simple_info, _simple_warn)
    from sphinx.util.osutil import ensuredir
    from sphinx.util.inspect import safe_getattr
    from jinja2 import FileSystemLoader, TemplateNotFound
    from jinja2.sandbox import SandboxedEnvironment

    from .utils import find_autosummary_in_lines_for_automodsumm as find_autosummary_in_lines

    if info is None:
        info = _simple_info
    if warn is None:
        warn = _simple_warn

    # info('[automodsumm] generating automodsumm for: ' + srcfn)

    # Create our own templating environment - here we use Astropy's
    # templates rather than the default autosummary templates, in order to
    # allow docstrings to be shown for methods.
    template_dirs = [os.path.join(os.path.dirname(__file__), 'templates'),
                     os.path.join(base_path, '_templates')]
    if builder is not None:
        # allow the user to override the templates
        template_loader = BuiltinTemplateLoader()
        template_loader.init(builder, dirs=template_dirs)
    else:
        if template_dir:
            template_dirs.insert(0, template_dir)
        template_loader = FileSystemLoader(template_dirs)
    template_env = SandboxedEnvironment(loader=template_loader)

    # read
    # items = find_autosummary_in_files(sources)
    items = find_autosummary_in_lines(lines, filename=srcfn)
    if len(items) > 0:
        msg = '[automodsumm] {1}: found {0} automodsumm entries to generate'
        info(msg.format(len(items), srcfn))

#    gennms = [item[0] for item in items]
#    if len(gennms) > 20:
#        gennms = gennms[:10] + ['...'] + gennms[-10:]
#    info('[automodsumm] generating autosummary for: ' + ', '.join(gennms))

    # remove possible duplicates
    items = list(set(items))

    # keep track of new files
    new_files = []

    # write
    for name, path, template_name, inherited_mem in sorted(items):

        if path is None:
            # The corresponding autosummary:: directive did not have
            # a :toctree: option
            continue

        path = os.path.abspath(os.path.join(base_path, path))
        ensuredir(path)

        try:
            import_by_name_values = import_by_name(name)
        except ImportError as e:
            warn('[automodsumm] failed to import %r: %s' % (name, e))
            continue

        # if block to accommodate Sphinx's v1.2.2 and v1.2.3 respectively
        if len(import_by_name_values) == 3:
            name, obj, parent = import_by_name_values
        elif len(import_by_name_values) == 4:
            name, obj, parent, module_name = import_by_name_values

        fn = os.path.join(path, name + suffix)

        # skip it if it exists
        if os.path.isfile(fn):
            continue

        new_files.append(fn)

        f = open(fn, 'w')

        try:

            if SPHINX_LT_17:
                doc = get_documenter(obj, parent)
            else:
                doc = get_documenter(app, obj, parent)

            if template_name is not None:
                template = template_env.get_template(template_name)
            else:
                tmplstr = 'autosummary_core/%s.rst'
                try:
                    template = template_env.get_template(tmplstr % doc.objtype)
                except TemplateNotFound:
                    template = template_env.get_template(tmplstr % 'base')

            def get_members_mod(obj, typ, include_public=[]):
                """
                typ = None -> all
                """
                items = []
                for name in dir(obj):
                    try:
                        if SPHINX_LT_17:
                            documenter = get_documenter(safe_getattr(obj, name), obj)
                        else:
                            documenter = get_documenter(app, safe_getattr(obj, name), obj)
                    except AttributeError:
                        continue
                    if typ is None or documenter.objtype == typ:
                        items.append(name)
                public = [x for x in items
                          if x in include_public or not x.startswith('_')]
                return public, items

            def get_members_class(obj, typ, include_public=[],
                                  include_base=False):
                """
                typ = None -> all
                include_base -> include attrs that are from a base class
                """
                items = []

                # using dir gets all of the attributes, including the elements
                # from the base class, otherwise use __slots__ or __dict__
                if include_base:
                    names = dir(obj)
                else:
                    if hasattr(obj, '__slots__'):
                        names = tuple(getattr(obj, '__slots__'))
                    else:
                        names = getattr(obj, '__dict__').keys()

                for name in names:
                    try:
                        if SPHINX_LT_17:
                            documenter = get_documenter(safe_getattr(obj, name), obj)
                        else:
                            documenter = get_documenter(app, safe_getattr(obj, name), obj)
                    except AttributeError:
                        continue
                    if typ is None or documenter.objtype == typ:
                        items.append(name)
                public = [x for x in items
                          if x in include_public or not x.startswith('_')]
                return public, items

            ns = {}

            if doc.objtype == 'module':
                ns['members'] = get_members_mod(obj, None)
                ns['functions'], ns['all_functions'] = \
                    get_members_mod(obj, 'function')
                ns['classes'], ns['all_classes'] = \
                    get_members_mod(obj, 'class')
                ns['exceptions'], ns['all_exceptions'] = \
                    get_members_mod(obj, 'exception')
            elif doc.objtype == 'class':
                if inherited_mem is not None:
                    # option set in this specifc directive
                    include_base = inherited_mem
                else:
                    # use default value
                    include_base = inherited_members

                api_class_methods = ['__init__', '__call__']
                ns['members'] = get_members_class(obj, None,
                                                  include_base=include_base)
                ns['methods'], ns['all_methods'] = \
                    get_members_class(obj, 'method', api_class_methods,
                                      include_base=include_base)
                ns['attributes'], ns['all_attributes'] = \
                    get_members_class(obj, 'attribute',
                                      include_base=include_base)
                ns['methods'].sort()
                ns['attributes'].sort()

            parts = name.split('.')
            if doc.objtype in ('method', 'attribute'):
                mod_name = '.'.join(parts[:-2])
                cls_name = parts[-2]
                obj_name = '.'.join(parts[-2:])
                ns['class'] = cls_name
            else:
                mod_name, obj_name = '.'.join(parts[:-1]), parts[-1]

            ns['fullname'] = name
            ns['module'] = mod_name
            ns['objname'] = obj_name
            ns['name'] = parts[-1]

            ns['objtype'] = doc.objtype
            ns['underline'] = len(obj_name) * '='

            # We now check whether a file for reference footnotes exists for
            # the module being documented. We first check if the
            # current module is a file or a directory, as this will give a
            # different path for the reference file. For example, if
            # documenting astropy.wcs then the reference file is at
            # ../wcs/references.txt, while if we are documenting
            # astropy.config.logging_helper (which is at
            # astropy/config/logging_helper.py) then the reference file is set
            # to ../config/references.txt
            if '.' in mod_name:
                mod_name_dir = mod_name.replace('.', '/').split('/', 1)[1]
            else:
                mod_name_dir = mod_name
            if not os.path.isdir(os.path.join(base_path, mod_name_dir)) \
               and os.path.isdir(os.path.join(base_path, mod_name_dir.rsplit('/', 1)[0])):
                mod_name_dir = mod_name_dir.rsplit('/', 1)[0]

            # We then have to check whether it exists, and if so, we pass it
            # to the template.
            if os.path.exists(os.path.join(base_path, mod_name_dir, 'references.txt')):
                # An important subtlety here is that the path we pass in has
                # to be relative to the file being generated, so we have to
                # figure out the right number of '..'s
                ndirsback = path.replace(base_path, '').count('/')
                ref_file_rel_segments = ['..'] * ndirsback
                ref_file_rel_segments.append(mod_name_dir)
                ref_file_rel_segments.append('references.txt')
                ns['referencefile'] = os.path.join(*ref_file_rel_segments)

            rendered = template.render(**ns)
            f.write(cleanup_whitespace(rendered))
        finally:
            f.close()


def setup(app):

    # need autodoc fixes
    # Note: we use __name__ here instead of just writing the module name in
    #       case this extension is bundled into another package
    from . import autodoc_enhancements
    app.setup_extension(autodoc_enhancements.__name__)

    # need inheritance-diagram for automod-diagram
    app.setup_extension('sphinx.ext.inheritance_diagram')

    app.add_directive('automod-diagram', Automoddiagram)
    app.add_directive('automodsumm', Automodsumm)
    app.connect('builder-inited', process_automodsumm_generation)

    app.add_config_value('automodsumm_writereprocessed', False, True)
    app.add_config_value('automodsumm_inherited_members', False, 'env')



import pytest
import numpy as np

from ..dataReduction import compute_transmissions
from ..skySurvey import directory


# Set up the random number generator
np.random.seed(1234)

import os.path
cal_directory = os.path.join(directory, "tests/test_data/test_dataReduction/ha")

def test_basic():
	"""
	Basic tuner spectrum load
	"""
	from ..skySurvey import SkySurvey
	file_list = os.path.join(cal_directory, "lamp_H-Alpha_45.0_65.2_203312_spec.fts")
	spec = SkySurvey(tuner_list = file_list)
	spec2 = SkySurvey(tuner_list = [file_list])

	assert np.allclose(spec["DATA"], spec2["DATA"], equal_nan = True)

def test_iter_error():
	"""
	Ensure tuner_list is iterable
	"""
	from ..skySurvey import SkySurvey
	file_list = 0
	try:
		spec = SkySurvey(tuner_list = file_list)
	except TypeError:
		assert True
	else:
		assert False







import logging

from astropy import units as u
from astropy.coordinates import Angle

import numpy as np 

from scipy.interpolate import interp1d

try:
    from spectral_cube import SpectralCube
except ModuleNotFoundError:
    # Error handling
    pass


import os.path 

directory = os.path.dirname(__file__)

track_dict = {
    "3kf":"3kF_lbvRBD.dat", "3kpc_far":"3kF_lbvRBD.dat", 
    "3kn":"3kN_lbvRBD.dat", "3kpc_near":"3kN_lbvRBD.dat", 
    "aqr":"AqR_lbvRBD.dat", "aquila_rift":"AqR_lbvRBD.dat", 
    "aqs":"AqS_lbvRBD.dat", "aquila_spur":"AqS_lbvRBD.dat", 
    "cnn":"CnN_lbvRBD.dat", "connecting_near":"CnN_lbvRBD.dat", 
    "cnx":"CnX_lbvRBD.dat", "connecting_extension":"CnX_lbvRBD.dat", 
    "crn":"CrN_lbvRBD.dat", "carina_near":"CrN_lbvRBD.dat", 
    "crf":"CrF_lbvRBD.dat", "carina_far":"CrF_lbvRBD.dat", 
    "ctn":"CtN_lbvRBD.dat", "centaurus_near":"CtN_lbvRBD.dat", 
    "ctf":"CtF_lbvRBD.dat", "centaurus_far":"CtF_lbvRBD.dat", 
    "loc":"Loc_lbvRBD.dat", "local":"Loc_lbvRBD.dat", 
    "los":"LoS_lbvRBD.dat", "local_spur":"LoS_lbvRBD.dat", 
    "nor":"Nor_lbvRBD.dat", "norma":"Nor_lbvRBD.dat", 
    "osc":"OSC_lbvRBD.dat", "outer_centaurus":"OSC_lbvRBD.dat", 
        "outer_scutum":"OSC_lbvRBD.dat", "outer_scutum_centaurus":"OSC_lbvRBD.dat", 
    "outer":"Out_lbvRBD.dat", "out":"Out_lbvRBD.dat", 
    "per":"Per_lbvRBD.dat", "perseus":"Per_lbvRBD.dat", 
    "scn":"ScN_lbvRBD.dat", "scutum_near":"ScN_lbvRBD.dat", 
    "scf":"ScF_lbvRBD.dat", "scutum_far":"ScF_lbvRBD.dat", 
    "sgn":"SgN_lbvRBD.dat", "sagittarius_near":"SgN_lbvRBD.dat",
    "sgf":"SgF_lbvRBD.dat", "sagittarius_far":"SgF_lbvRBD.dat"
}

def get_lbv_track(filename = None, reid_track = None, **kwargs):
    """
    returns (longitude, latitude, velocity, radius, beta, distance) for spiral arm tracks

    Parameters
    ----------
    filename:'str', optional, must be keyword
        Filename of custom LBV track to be read in
    reid_track:'str', optional, must be keyword
        Spiral Arm track from Reid et al. (2016) to read in
        Options are not case sensitive:
            3kpc Arm:
                Far: "3kf", "3kpc_far"
                Near: "3kn", "3kpc_near"
            Aquila Rift: "AqR", "Aquila_Rift"
            Aquila Spur: "AqS", "Aquila_Spur"
            Connecting Arm:
                Near: "CnN", "Connecting_Near"
                Extension: "CnX", "Connecting_Extension"
            Carina Arm:
                Near: "CrN", "Carina_Near"
                Far: "CrF", "Carina_Far"
            Centaurus Arm:
                Near: "CtN", "Centaurus_Near"
                Far: "CtF", "Centaurus_Far"
            Local Arm:
                "Loc", "Local"
                Spur?: "LoS", "Local_Spur"
            Norma Arm:
                "Nor", "Norma"
            Outer Scutum Centaurus Arm:
                "OSC", "Outer_Scutum", "Outer_Centaurus", "Outer_Scutum_Centaurus"
            Outer Arm:
                "Outer", "Out"
            Perseus Arm:
                "Per", "Perseus"
            Scutum Arm:
                Near: "ScN", "Scutum_Near"
                Far: "ScF", "Scutum_Far"
            Sagittarius Arm:
                Near: "SgN", "Sagittarius_Near"
                Far: "SgF", "Sagittarius_Far"
    kwargs:
        optional keywords passed to np.genfromtxt
    """
    # Check for Reid Track Match:
    if reid_track is not None:
        try:
            filename = os.path.join(directory, "data", "Reid16_SpiralArms", track_dict[reid_track.lower()])
        except KeyError:
            raise KeyError("Reid et al. (2016) Spiral Arm track for {} was not found!".format(reid_track))
    else: 
        logging.warning("Not using provided data files - expecting columns to be ordered as lbv_RBD.")

    # Try reading in data:
    if not "comments" in kwargs:
        # Default comments flag in Reid lbv_RBD Tracks
        kwargs["comments"] = "!"

    track = np.genfromtxt(filename, **kwargs)

    # Check shape
    if track.shape[1] != 6:
        logging.warning("Track file did not have expected number of columns - adding in columns of 0s")
        new_track = np.zeros((track.shape[0], 6))
        for column in range(track.shape[1]):
            new_track[:,column] = track[:,column]

        track = new_track

    return track


def get_spiral_slice(survey, track = None, filename = None, 
    brange = None, lrange = None, 
    interpolate = True, wrap_at_180 = False, 
    vel_width = None, return_track = False):
    """
    Returns SkySurvey object isolated to velocity ranges corresponding to specified spiral arm

    Parameters
    ----------
    survey: 'whampy.skySurvey'
        input skySurvey
    track: 'np.ndarray', 'str', optional, must be keyword
        if 'numpy array', lbv_RBD track data
        if 'str', name of spiral arm from Reid et al. (2016)
    filename: 'str', optional, must be keyword
        filename of track file to read in using whampy.lbvTracks.get_lbv_track
    brange: 'list', 'u.Quantity'
        min,max latitude to restrict data to
        Default of +/- 40 deg
    lrange: 'list', 'u.Quantity'
        min,max longitude to restrict data to
        Default of full track extent
    interpolate: 'bool', optional, must be keyword
        if True, interpolates velocity to coordinate of pointings
        if False, ... do nothing ... for now
            Future: slices have velocities set by track 
    wrap_at_180: `bool`, optional, must be keyword
            if True, wraps longitude angles at 180d
            use if mapping accross Galactic Center
    vel_width: `number`, `u.Quantity`, optional, must be keyword
        velocity width to isolate in km/s
    return_track: `bool`, optional, must be keyword
        if True, will also return the spiral arm track as the second element of a tuple
    """
    # Check Track Data
    if track is not None:
        if not track.__class__ is np.ndarray:
            track = get_lbv_track(reid_track = track)
        else:
            if track.shape[1] < 2:
                raise TypeError("track should have at least two columns (l,v) or up to 6 (lbvRBD).")
    elif filename is None:
        raise SyntaxError("No track or track filename provided!")
    else:
        track = get_lbv_track(filename = filename)

    if wrap_at_180:
        wrap_at = "180d"
    else:
        wrap_at = "360d"

    # Check velocity width
    if vel_width is None:
        logging.warning("No Velocity width specified, using default of 16 km/s.")
        vel_width = 16*u.km/u.s
    elif not isinstance(vel_width, u.Quantity):
        vel_width *= u.km/u.s
        logging.warning("No units specified for vel_width, assuming km/s.")

    #Extract Needed track informattion
    if track.shape[1] == 2:
        lon_track = track[:,0] * u.deg
        lon_track = Angle(lon_track).wrap_at(wrap_at)
        vel_track = track[:,1] * u.km/u.s
    else: # Reid et al. (2016) format
        lon_track = track[:,0] * u.deg
        lon_track = Angle(lon_track).wrap_at(wrap_at)
        vel_track = track[:,2] * u.km/u.s


    # Set Default latitude range
    if brange is None:
        brange = [-40,40]*u.deg
    elif isinstance(brange, u.Quantity):
        brange = brange.to(u.deg)
    else: 
        brange = brange * u.deg
        logging.warning("No units specified for latitude range, assuming Degrees!")

    # Set Longitude Range
    if lrange is None:
        lrange = [np.min(lon_track.value), np.max(lon_track.value)]*u.deg
    elif isinstance(lrange, u.Quantity):
        lrange = lrange.to(u.deg)
    else: 
        lrange = lrange * u.deg
        logging.warning("No units specified for longitude range, assuming Degrees!")

    # Extract Relevant sky section
    bounds = [lrange[0], lrange[1], brange[0], brange[1]]
    sky_cut = survey.sky_section(bounds, wrap_at_180 = wrap_at_180)

    # Restrict Velocity Space with a mask
    wham_coords = sky_cut.get_SkyCoord()

    
    if interpolate:
        l_v_track = interp1d(lon_track, vel_track)
        central_velocities = l_v_track(wham_coords.l.wrap_at(wrap_at)) * vel_track.unit
    else:   ## For now - always interpolate
        l_v_track = interp1d(lon_track, vel_track)
        central_velocities = l_v_track(wham_coords.l.wrap_at(wrap_at)) * vel_track.unit

    # Get velocity masks
    # Is there a better way to do this without a loop?
    sky_cut["VEL_MASK"] = np.empty((len(sky_cut), len(sky_cut[0]["VELOCITY"])), dtype = bool)
    for ell, vel in enumerate(central_velocities):
        mask = sky_cut[ell]["VELOCITY"] <= (vel + vel_width/2).value
        mask &= sky_cut[ell]["VELOCITY"] >= (vel - vel_width/2).value
        sky_cut[ell]["VEL_MASK"] = mask

    # Set Intensities
    sky_cut["INTEN"], sky_cut["ERROR"] = sky_cut.moment(return_sigma = True, masked = True)


    # Save Track info as attribute
    sky_cut.lbv_RBD_track = track

    if not return_track:
        return sky_cut
    else:
        return sky_cut, track















import logging

from astropy import units as u
import numpy as np 
import matplotlib.pyplot as plt

from matplotlib.colors import LogNorm
# from mpl_toolkits.basemap import Basemap
try:
    import cartopy.crs as ccrs
except ModuleNotFoundError:
    # Error handling
    pass
try:
    from spectral_cube import SpectralCube
except ModuleNotFoundError:
    # Error handling
    pass

from astropy.coordinates import SkyCoord
from astropy.coordinates import Angle

from .clickMap import SpectrumPlotter

from .lbvTracks import get_spiral_slice
from .scaleHeight import get_scale_height_data
from .spectralStack import stack_spectra_bootstrap

from scipy.interpolate import griddata






class SkySurveyMixin(object):
    """
    Mixin class with convenience functions for WHAM data
    """

    def get_SkyCoord(self, **kwargs):
        """
        returns SkyCoord object of coordinates for pointings in Table

        Parameters
        ----------

        **kwargs:
            passed to SkyCoord
        """
        return SkyCoord(l = self["GAL-LON"], b = self["GAL-LAT"], frame = "galactic", **kwargs)

    def get_spectrum(self, coordinate, index = False):
        """
        returns single WHAM pointing closest to provided coordinate

        Parameters
        ----------

        coordinate: 'SkyCoord' or 'list'
            SkyCoord or [lon,lat] list
            assumes Galactic Coordinates and u.deg
        index: 'bool', optional, must be keyword
            if True, returns the index of the closest coordinate 

        """
        if not isinstance(coordinate, SkyCoord):
            if not isinstance(coordinate, u.Quantity):
                logging.warning("No units provided for coordinate, assuming u.deg")
                coordinate *= u.deg
            coordinate = SkyCoord(l = coordinate[0], b = coordinate[1], frame = 'galactic')

        # Ensure coordinate is in proper frame
        coordinate = coordinate.transform_to('galactic')

        closest_ind = coordinate.separation(self.get_SkyCoord()).argmin()

        if index:
            return closest_ind
        else:
            return self[closest_ind]

    def get_spectral_slab(self, vmin, vmax):
        """
        returns survey with velocity window cut down

        Parameters
        ----------

        vmin: `number`, `u.Quantity`
            Min velocity
        vmax: `number`, `u.Quantity`
            Max velocity
        """
        if hasattr(vmin, "unit"):
            vmin = vmin.to(u.km/u.s).value
        if hasattr(vmax, "unit"):
            vmax = vmax.to(u.km/u.s).value

        velocity_cut_mask = self["VELOCITY"].data >= vmin
        velocity_cut_mask &= self["VELOCITY"].data <= vmax

        if velocity_cut_mask.sum() == 0:
            raise ValueError("No Data within specified velocity range!")
        else:
            masked_velocity_entries = np.sum(velocity_cut_mask, axis = 0, dtype = bool)

            new_velocity_column_full = np.copy(self["VELOCITY"])
            new_velocity_column_full[~velocity_cut_mask] = np.nan
            new_velocity_column = new_velocity_column_full[:,masked_velocity_entries]

            new_data_column_full = np.copy(self["DATA"]) * self["DATA"].unit
            new_data_column = new_data_column_full[:,masked_velocity_entries]

            new_variance_column_full = np.copy(self["VARIANCE"]) * self["VARIANCE"].unit
            new_variance_column = new_variance_column_full[:,masked_velocity_entries]

            self.remove_column("VELOCITY")
            self["VELOCITY"] = new_velocity_column * u.km/u.s

            self.remove_column("VARIANCE")
            self["VARIANCE"] = new_variance_column

            self.remove_column("DATA")
            self["DATA"] = new_data_column


    def moment(self, order = None, vmin = None, vmax = None, 
        return_sigma = False, masked = False):
        """
        compute moment maps 

        Parameters
        ----------
        order: 'number', optional, must be keyword
            moment order to return, default = 0
        vmin: 'number' or 'Quantuty', optional, must be keyword
            min Velocity, default units of km/s
        vmax: 'number' or 'Quantity', optional, must be keyword
            max Velocity, default units of km/s
        return_sigma: 'bool', optional, must be keyword
            if True, will also return one-sigma gaussian error estimate
        masked: `bool`, optional, must be keyword
            if True, used masked velocity axis
        """

        if order is None:
            order = 0 # Assume default value

        # Mask out nan values
        nan_msk = np.isnan(self["DATA"]) | np.isnan(self["VELOCITY"])
        # Mask out negative data values
        nan_msk |= self["DATA"] < 0.

        # masked velocity axis
        if masked:
            nan_msk |= np.invert(self["VEL_MASK"])

        if return_sigma:
            nan_msk |= np.isnan(self["VARIANCE"])

        # Velocity mask if applicable:
        if vmin is not None:
            if not isinstance(vmin, u.Quantity):
                logging.warning("No units specified for vmin, assuming u.km/u.s")
                vmin *= u.km/u.s

            nan_msk |= self["VELOCITY"] <= vmin.to(u.km/u.s).value

        if vmax is not None:
            if not isinstance(vmax, u.Quantity):
                logging.warning("No units specified for vmax, assuming u.km/u.s")
                vmax *= u.km/u.s

            nan_msk |= self["VELOCITY"] >= vmax.to(u.km/u.s).value

        data_masked = np.ma.masked_array(self["DATA"], mask = nan_msk)
        vel_masked = np.ma.masked_array(self["VELOCITY"], mask = nan_msk)
        var_masked = np.ma.masked_array(self["VARIANCE"], mask = nan_msk)

        # Zeroth Order Moment
        moment_0 = np.trapz(data_masked, x = vel_masked, 
            axis = 1) * self["DATA"].unit * self["VELOCITY"].unit
        err_0 = np.trapz(np.sqrt(var_masked), x = vel_masked, axis = 1) * self["DATA"].unit * self["VELOCITY"].unit

        if order > 0:
            moment_1 = np.trapz(data_masked * vel_masked, x = vel_masked, 
                axis = 1) * self["DATA"].unit * self["VELOCITY"].unit**2 / moment_0
            err_1_subover_mom_1 = np.trapz(np.sqrt(var_masked) * vel_masked**2, x = vel_masked, 
                            axis = 1) * self["DATA"].unit * self["VELOCITY"].unit**2 / (moment_1 * moment_0)
            err_1 = moment_1 * np.sqrt(err_1_subover_mom_1**2 + (err_0 / moment_0)**2)
            if order > 1:
                moment_2 = np.trapz(data_masked * (vel_masked - moment_1.value[:,None])**2, 
                    x = vel_masked, 
                    axis = 1) * self["DATA"].unit * self["VELOCITY"].unit**3 / moment_0
                print(moment_2.unit)
                err_2_subover_mom2 = np.trapz(data_masked * (vel_masked - moment_1.value[:,None])**2 * np.sqrt(var_masked / 
                                data_masked**2 + 2*(err_1[:,None] / moment_1[:,None])**2), 
                    x = vel_masked, 
                    axis = 1) * self["DATA"].unit * self["VELOCITY"].unit**3 / (moment_2 * moment_0)
                print(err_2_subover_mom2.unit)
                err_2 = moment_2 * np.sqrt(err_2_subover_mom2**2 + (err_0 / moment_0)**2)
                if return_sigma:
                    return moment_2, err_2
                else:
                    return moment_2
            else:
                if return_sigma:
                    return moment_1, err_1
                else:
                    return moment_1
        else:
            if return_sigma:
                return moment_0, err_0
            else:
                return moment_0


    # Plotting Functions
    def intensity_map(self, fig = None, ax = None, lrange = None, brange = None, 
                        vel_range = None,
                        s_factor = 1., colorbar = False, cbar_kwargs = {}, 
                        return_sc = False, 
                        smooth = False, smooth_res = None, **kwargs):
        """
        plots the intensity map using WHAM beams as scatter plots

        Parameters
        -----------
        
        fig: 'plt.figure', optional, must be keyword
            if provided, will create axes on the figure provided
        ax: 'plt.figure.axes' or 'cartopy.axes`, optional, must be keyword
            if provided, will plot on these axes
            can provide ax as a cartpy projection that contains different map projections
        lrange: 'list', optional, must be keyword
            provides longitude range to plot
        brange: 'list', optional, must be keyword
            provides latitude range to plot
        vel_range: 'list' or 'Quantity', optional, must be keyword
            velocity range to integrate data over
        s_factor: 'number', optional, must be keyword
            multiplied by supplied default s value to set size of WHAM beams
        colorbar: 'bool', optional, must be keyword
            if True, plots colorbar
        cbar_kwargs: 'dict', optional, must be keyword
            dictionary of kwargs to pass to colorbar
        return_sc: 'bool', optional, must be keyword
            if True, returns ax.scatter object before figure
        smooth: `bool`, optional, must be keyword
            if True, smooths map using griddata and plots using pcolormesh
        smooth_res: `number`, optional, must be keyword
            pixel width in units of degrees
        **kwarrgs: `dict`
            passed to scatter plot

        """
        if not hasattr(ax, 'scatter'):
            if not hasattr(fig, 'add_subplot'):
                fig = plt.figure()
                ax = fig.add_subplot(111)
            else:
                ax = fig.add_subplot(111)
        else:
            if not hasattr(fig, 'add_subplot'):
                fig = plt.gcf()


        wham_coords = self.get_SkyCoord()

        if lrange is None:
            lrange_s = [wham_coords.l.wrap_at("180d").max().value, wham_coords.l.wrap_at("180d").min().value]
        elif isinstance(lrange, u.Quantity):
            lrange = Angle(lrange).wrap_at("180d").value
        else:
            logging.warning("No units provided for lrange, assuming u.deg")
            lrange = Angle(lrange*u.deg).wrap_at("180d").value
        if brange is None:
            brange_s = [wham_coords.b.min().value, wham_coords.b.max().value]
        elif isinstance(brange, u.Quantity):
            brange = brange.to(u.deg).value


        if not smooth:
            if not "s" in kwargs:
                size = fig.get_size_inches()*fig.dpi
                if brange is not None:
                    brange_s = brange
                if lrange is not None:
                    lrange_s = lrange
                s = np.min([size / np.abs(np.diff(lrange_s)), size / np.abs(np.diff(brange_s))]) * s_factor
                kwargs["s"] = s

        if not "c" in kwargs:
            if vel_range is None:
                kwargs["c"] = self["INTEN"]
            elif not isinstance(vel_range, u.Quantity):
                logging.warning("No units provided for vel_range, assuming u.km/u.s")
                vel_range *= u.km/u.s
            else:
                vel_range = vel_range.to(u.km/u.s)
                kwargs["c"] = self.moment(order = 0, vmin = vel_range.min(), vmax = vel_range.max())

        if not "cmap" in kwargs:
            kwargs["cmap"] = 'plasma'

        if not "vmin" in kwargs:
            kwargs["vmin"] = 0.1

        if not "vmax" in kwargs:
            kwargs["vmax"] = 100.

        if not "norm" in kwargs:
            kwargs["norm"] = LogNorm()
        if hasattr(ax, "coastlines"):
            if not "transform" in kwargs:
                kwargs["transform"] = ccrs.PlateCarree()
                print("No transform specified with cartopy axes projection, assuming PlateCarree")

        lon_points = wham_coords.l.wrap_at("180d")
        lat_points = wham_coords.b.wrap_at("180d")

        if smooth:
            if smooth_res is None:
                smooth_res = 0.2
            if lrange is None:
                lrange = lrange_s
            if brange is None:
                brange = brange_s
            if lrange[1] < lrange[0]:
                gridx = np.flip(np.arange(lrange[1], lrange[0] + smooth_res, smooth_res))
            else:
                gridx = np.arange(lrange[0], lrange[1] + smooth_res, smooth_res)
            gridy = np.arange(brange[0], brange[1] + smooth_res, smooth_res)

            zi = griddata((lon_points, lat_points), kwargs["c"],
                (gridx[None,:], gridy[:,None]), 
                method='cubic')
            
            c_kwarg = kwargs["c"]
            del kwargs["c"]

        if hasattr(ax, "wcs"):
            if ax.wcs.naxis == 3:
                if smooth:
                    gridx, gridy, _ = ax.wcs.wcs_world2pix(gridx, gridy, np.zeros_like(gridx), 0)
                else:
                    lon_points, lat_points, _ = ax.wcs.wcs_world2pix(lon_points, lat_points, np.zeros_like(lon_points.value), 0)
            elif ax.wcs.naxis == 2:
                if smooth:
                    gridx, gridy = ax.wcs.wcs_world2pix(gridx, gridy, 0)
                else:
                    lon_points, lat_points = ax.wcs.wcs_world2pix(lon_points, lat_points, 0)



        # Plot the WHAM beams
        if smooth:
            sc = ax.pcolormesh(gridx, gridy, zi, **kwargs)
        else:
            sc = ax.scatter(lon_points, lat_points, **kwargs)

        if not hasattr(ax, "coastlines"):
            if (lrange is not None) & (brange is not None):
                ax.set_xlim(lrange)
                ax.set_ylim(brange)
            else:
                ax.invert_xaxis()
            ax.set_xlabel("Galactic Longitude (deg)", fontsize = 12)
            ax.set_ylabel("Galactic Latitude (deg)", fontsize = 12)
        else:
            ax.invert_xaxis()
            if (lrange is not None) & (brange is not None):
                ax.set_extent([lrange[0], lrange[1], brange[0], brange[1]])   
            try:
                ax.gridlines(draw_labels = True)
            except TypeError:
                ax.gridlines()



        if colorbar:
            if not "label" in cbar_kwargs:
                if "c" in kwargs:
                    cbar_kwargs["label"] = "H-Alpha Intensity ({})".format(kwargs["c"].unit)
                else:
                    cbar_kwargs["label"] = "H-Alpha Intensity ({})".format(c_kwarg.unit)


            cb = plt.colorbar(sc, **cbar_kwargs)

        if return_sc:
            return sc, fig
        else:
            return fig

    def click_map(self, fig = None, image_ax = None, spec_ax = None, 
                    projection = None, spectra_kwargs = {}, 
                    over_data = None, average_beam = False, 
                    radius = None, over_spectra_kwargs = {}, 
                    over_spec_ax = None, share_yaxis = False,
                     **kwargs):
                    
        """
        Interactive plotting of WHAM data to plot spectra when clicking on map

        Parameters
        ----------

        fig: 'plt.figure', optional, must be keyword
            if provided, will create axes on the figure provided
        image_ax: 'plt.figure.axes' or 'cartopy.axes`, optional, must be keyword
            if provided, will plot image/map on these axes
            can provide ax as a cartpy projection that contains different map projections 
        spec_ax: 'plt.figure.axes', optional, must be keyword
            if provided, will plot spectra on these axes
        projection:   'ccrs.projection'
            if provided, will be passed to creating a map with specified cartopy projection
        spectra_kwargs: 'dict', optional, must be keyword
            kwargs passed to plot command for spectra
        over_data:  'SkySurvey' or 'str' or 'spectral_cube.SpectralCube', optional, must be keyword
            Extra data to over plot spectra from
            if SkySurvey, assumed to be WHAM observations, perhaps at another wavelength
            if 'str', assumed to be a 3D FITS Data cube filename to be loaded as SpectralCube
            if 'SpectralCube', defaults to extracting closest spectra to click
        average_beam: 'bool', optional, must be keyword
            if True, instead over plots average spectrum from over_data within nearest WHAM beam
        radius: 'Quantity' or  'number', optional, must be keyword
            beam radius to average beam over if avreage_beam is True
            default is 0.5 degrees (WHAM beam)
        over_spectra_kwargs: 'dict', optional, must be keyword
            kwargs passed to plot command for over_spectra
        over_spec_ax: 'plt.figure.axes', optional, must be keyword
            if provided, will plot over_spectra on these axes
        share_yaxis: 'bool', optional, must be keyword
            if True, over_spectra shares same y_axis as spectra
            if False, over_spec_ax has a unique y_axis
            if over_spec_ax is provided, share_yaxis is not used
        **kwargs: 'dict', must be keywords
            passed to `SkySurvey.intensity_map`
        """
        if not hasattr(image_ax, 'scatter'):
            if not hasattr(fig, 'add_subplot'):
                # Create Figure and subplot for image
                fig = plt.figure()
            if projection is not None:
                image_ax = fig.add_subplot(111, projection = projection)
            else:
                image_ax = fig.add_subplot(111)
        if not hasattr(spec_ax, 'scatter'):
            # make room for spectra at bottom
            fig.subplots_adjust(bottom = 0.5)
            spec_ax = fig.add_axes([0.1, .1, .8, .3])
        if not hasattr(over_spec_ax, 'scatter'):
            if over_data is not None:
                if share_yaxis:
                    # Same axis for both spectra
                    over_spec_ax = spec_ax
                else:
                    # Shared x axis but unique y axes
                    over_spec_ax = spec_ax.twinx()


        # Plot image
        fig = self.intensity_map(fig = fig, ax = image_ax, **kwargs)

        # Check for and assign default plot parameters
        if ("lw" not in spectra_kwargs) & ("linewidth" not in spectra_kwargs):
            spectra_kwargs["lw"] = 2
        if ("c" not in spectra_kwargs) & ("color" not in spectra_kwargs):
            spectra_kwargs["c"] = 'red'
        if ("ls" not in spectra_kwargs) & ("linestyle" not in spectra_kwargs):
            spectra_kwargs["ls"] = '-'



        # Empty line
        spec, = spec_ax.plot([0],[0], **spectra_kwargs)

        # Over Plot extra Spectra if needed
        if over_data is not None:
            if ("lw" not in over_spectra_kwargs) & ("linewidth" not in over_spectra_kwargs):
                over_spectra_kwargs["lw"] = 1
            if ("c" not in over_spectra_kwargs) & ("color" not in over_spectra_kwargs):
                over_spectra_kwargs["c"] = 'blue'
            if ("ls" not in over_spectra_kwargs) & ("linestyle" not in over_spectra_kwargs):
                over_spectra_kwargs["ls"] = '--'

            # Empty line 
            over_spec, = over_spec_ax.plot([0],[0], **over_spectra_kwargs)
        else:
            over_spec = None

        return SpectrumPlotter(image_ax, spec, 
                                data = self, 
                                over_data = over_data, 
                                over_line = over_spec, 
                                average_beam = average_beam, 
                                radius = radius)

    def get_spiral_slice(self, **kwargs):
        """
        Returns SkySurvey object isolated to velocity ranges corresponding to specified spiral arm

        Parameters
        ----------
        track: 'np.ndarray', 'str', optional, must be keyword
            if 'numpy array', lbv_RBD track data
            if 'str', name of spiral arm from Reid et al. (2016)
        filename: 'str', optional, must be keyword
            filename of track file to read in using whampy.lbvTracks.get_lbv_track
        brange: 'list', 'u.Quantity'
            min,max latitude to restrict data to
            Default of +/- 40 deg
        lrange: 'list', 'u.Quantity'
            min,max longitude to restrict data to
            Default of full track extent
        interpolate: 'bool', optional, must be keyword
            if True, interpolates velocity to coordinate of pointings
            if False, ... do nothing ... for now
                Future: slices have velocities set by track 
        wrap_at_180: `bool`, optional, must be keyword
                if True, wraps longitude angles at 180d
                use if mapping accross Galactic Center
        vel_width: `number`, `u.Quantity`, optional, must be keyword
            velocity width to isolate in km/s
        """
        return get_spiral_slice(self, **kwargs)

    def get_scale_height_data(self, **kwargs):
        """
        Return data needed for scale height analysis
        
        Parameters
        ----------
        data: `skySurvey`
            WHAM skySurvey object of full sky (requires track keyword), or spiral arm section
        track: `str`, optional, must be keyword
            if provided, will apply skySurvey.get_spiral_slice for provided track
            if None, will check for track as an attribute of data
        deredden: `bool`, `dustmap`, optional, must be keyword
            if True, will apply dereddening using 3D dustmaps of Marshall et al. (2006)
            or can input a dustmap to query from using the dustmaps package
            default to `dustmaps.marshall.MarshallQuery`
            Warning: Currently only supports Marshall Dustmap
        return_pandas_dataframe: `bool`, optional, must be keyword
            if True, returns pandas dataframe with subset of data specific to scale height analysis
        **kwargs: `dict`
            keywords passed to data.get_spiral_slice if track is provided
        """
        return get_scale_height_data(self, **kwargs)

    def stack_spectra_bootstrap(self, **kwargs):
        """
        Stack all spectra in provided SkySurvey Table
        
        Parameters
        ----------
        survey: 'whampy.skySurvey'
            input skySurvey
        data_column: 'str', optional, must be keyword
            Name of data column, default of "DATA"
        variance_column: 'str', optional, must be keyword
            Name od variance column, defaut of "VARIANCE"
        velocity: 'np.array, list', optional, must be keyword
            velocity to interpolate data to for stacking
            defaults to np.round of first row "VELOCITY"
        no_interp: 'bool', optional, must be keyword
            if True, skips interpolating data to same velocity
            assumes it is already regularly gridded to given velocity
        n_boot: 'int', optional, must be keyword
            number of bootstrap samples for each spectrum to stack
            default of 1000
        ci: 'number', optional, must be keyword
            Size of confidence interval to return as errors
        estimator: 'callable', optional, must be keyword
            estimator method to use, defaults to `numpy.mean`
            must take axis keyword
        set_name: 'str', optional, must be keyword
            if provided, sets "NAME" column to this in returned stack
        """
        return stack_spectra_bootstrap(self, **kwargs)

    def get_equivalent_width(self, 
                            intensity = None, 
                            intensity_error = None, 
                            continuum = None, 
                            continuum_error = None, 
                            return_sigma = True):
        """
        Estimate emission line equivalent width

        Parameters
        ----------
        continuum: `u.Quantity`, optional, must be keyword
            Continuum value or values to use
        continuum_error: `u.Quantity`, optional, must be keyword
            Continuum error value or values to use
        intensity: 'u.Quantity', optional, must be keyword
            Intensity values to use
        """
        # Equivalencies
        ha_wave = 656.3 * u.nm

        optical_ha_equiv = u.doppler_optical(ha_wave)

        # Chanel Width
        dv = (self["VELOCITY"][0][1] - self["VELOCITY"][0][0]) * u.km/u.s
        dlambda = (dv).to(u.AA, equivalencies=optical_ha_equiv) - \
            (0 * u.km/u.s).to(u.AA, equivalencies=optical_ha_equiv)

        if continuum is None:
            if "BKG" in self.keys():
                continuum = self["BKG"] / 22.8 * u.R / dlambda
                if "BKGSD" in self.keys():
                    continuum_error = self["BKGSD"] / 22.8 * u.R / dlambda
                elif continuum_error is None:
                    logging.warning("No Continuum Error Found, using 10% errors instead.")
                    continuum_error = 0.1 * continuum

            elif not hasattr(continuum, "unit"):
                raise TypeError("continuum must be Quantity with units flux/Angstrum")

        if intensity is None:
            ew = self["INTEN"] / continuum
            if return_sigma:
                if intensity_error is None: 
                    ew_error = np.sqrt((self["ERROR"].data / self["INTEN"].data)**2 + \
                                        (continuum_error / continuum)**2) * ew
                else:
                    ew_error = np.sqrt((intensity_error / self["INTEN"])**2 + \
                                        (continuum_error / continuum)**2) * ew
        else:
            ew = intensity / continuum
            if return_sigma:
                if intensity_error is None: 
                    raise TypeError("if intensity is provided, must also provide intensity_error")

                else:
                    ew_error = np.sqrt((intensity_error / intensity)**2 + \
                                        (continuum_error / continuum)**2) * ew

        if return_sigma:
            return ew, ew_error
        else:
            return ew



















import pytest
import matplotlib.pyplot as plt
import astropy.units as u
import numpy as np
from ..skySurvey import SkySurvey 

# Set up the random number generator.
np.random.seed(1234)

# Local directory:
# "/Users/dk/Data/WHAM/wham-ss-DR1-v161116-170912.fits"

survey = SkySurvey()

BASELINE_DIR = 'baseline'

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic():
    """
    Basic intensity plot
    """
    fig = survey.intensity_map()
    return fig

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_fig_input():
    """
    fig kwarg
    """
    fig = plt.figure()
    return survey.intensity_map(fig = fig)

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_ax_input():
    """
    ax kwarrg
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    return survey.intensity_map(ax = ax)

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_ranges():
    """
    lrange and brange
    """
    lrange = [30,-30]
    brange = [-30,30]
    return survey.intensity_map(lrange = lrange * u.deg, brange = brange)

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_ranges_unit():
    """
    lrange and brange
    """
    lrange = [30,-30]
    brange = [-30,30]
    return survey.intensity_map(lrange = lrange, brange = brange*u.deg)

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_colorbar():
    """
    colorbar
    """
    return survey.intensity_map(colorbar = True, 
        cbar_kwargs = {"orientation":"horizontal"})

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_sc():
    """
    return_sc test
    """
    sc, fig = survey.intensity_map(return_sc = True)
    cb = fig.colorbar(sc, orientation = "horizontal")
    return fig

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_cartopy_stereo():
    import cartopy.crs as ccrs
    """
    test stereo projection with cartopy
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection = ccrs.Stereographic())
    fig = survey.intensity_map(ax = ax, fig = fig, s_factor=2, 
        lrange = [40,-40], brange = [-20,20])
    return fig

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_vel_range():
    """
    test velocity range input
    """
    vel_range = [-20,20]
    return survey.intensity_map(vel_range = vel_range)

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_vel_range_units():
    """
    test velocity range input
    """
    vel_range = [-20,20]*u.km/u.s
    return survey.intensity_map(vel_range = vel_range)

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_smooth():
    """
    Basic intensity plot smooth
    """
    fig = survey.intensity_map(smooth = True)
    return fig
    
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_smooth_lrange():
    """
    Basic intensity plot smooth
    """
    fig = survey.intensity_map(smooth = True, lrange = [50,-50])
    return fig






import logging

from astropy import units as u
import numpy as np 
import matplotlib.pyplot as plt

from.skySurvey import SkySurvey

from scipy import stats
from astropy.io import fits
import glob
import os.path

def get_calibration_files(directory, lines = None, calibrator = None):
    """
    Gets lists of all calibration files to use
    
    Parameters
    ----------
    
    directory: `str`
        directory with combo spectra
    lines: `str`, listlike, optional, must be keyword
        lines to load
        if None, searches for all default lines
    calibrator: `str`, listlike, optional, must be keyword
        calibrator to load
        if None, searches for all default calibrators
    """
    
    # Check calibrators keyword
    if calibrator is None:
        calibrator = ["zeta_oph", "g300_0", "spica_hii", "l_ori", "g194_0"]
    elif isinstance(calibrator, str):
        calibrator = [calibrator]
    elif not hasattr(calibrator, "__iter__"):
        raise TypeError("Invalid calibrator Type, calibrator must be a string, list of strings, or None.")
        
    # Check lines keyword
    if lines is None:
        lines = ["ha", "hb", "nii_red", "sii", "oiii", "oi", "hei"]
    elif isinstance(lines, str):
        lines = [lines]
    elif not hasattr(lines, "__iter__"):
        raise TypeError("Invalid lines Type, lines must be a string, list of strings, or None.")
        
    
    # Get all filenames in the target directory
    filename_lines_dict = {}
    for line in lines:
        filename_dict = {}
        for calibrator_name in calibrator:
            files = glob.glob(os.path.join(directory, 
                                           line, 
                                           "combo", 
                                           "{}-*.fts".format(calibrator_name.lower())))
            if len(files) > 0:
                filename_dict[calibrator_name] = files
        if len(filename_dict) > 0:
            filename_lines_dict[line] = filename_dict
    
    
    return filename_lines_dict


def read_calibration_data(directory, lines = None, calibrator = None):
    """
    Reads in calibration data as a SkySurvey object
    
    Parameters
    ----------
    
    directory: `str`
        directory with combo spectra
    lines: `str`, listlike, optional, must be keyword
        lines to load
        if None, searches for all default lines
    calibrator: `str`, listlike, optional, must be keyword
        calibrator to load
        if None, searches for all default calibrators
    
    """
    file_lists = get_calibration_files(directory, lines = lines, calibrator = calibrator)
    
    data_dict = {}
    for line in file_lists.keys():
        data_dict[line] = {}
        for calibrator_name in file_lists[line].keys():
            try:
                data_dict[line][calibrator_name] = SkySurvey(file_list = file_lists[line][calibrator_name])
            except ValueError:
                logging.warning("No {0}/{1} calibrators have the right fits extension!".format(line, 
                                                                                               calibrator_name))
            
    return data_dict

def compute_transmissions(directory, lines = None, calibrator = None, plot = False, alpha = None):
    """
    Computes transmission curves (slopes) for a given directory
    
    Parameters
    ----------
    
    directory: `str`
        directory with combo spectra
    lines: `str`, listlike, optional, must be keyword
        lines to load
        if None, searches for all default lines
    calibrator: `str`, listlike, optional, must be keyword
        calibrator to load
        if None, searches for all default calibrators
    plot: `bool`, optional, must be keyword
        if True, plots all data and slopes
    alpha: float, optional
        Confidence degree between 0 and 1. 
        Default is 95% confidence. 
    """
    # Default confidence degree
    if alpha is None:
        alpha = 0.95
    
    # Read data
    data_dict = read_calibration_data(directory, lines = lines, calibrator = calibrator)
    
    
    # Compute Slopees
    transmissions = {}
    for line in data_dict.keys():
        transmissions[line] = {}
        for calibrator_name in data_dict[line].keys():
            # Check that multiple data points exist
            if len(data_dict[line][calibrator_name]) > 2:
                transmissions[line][calibrator_name] = {}
                fit_result = stats.theilslopes(np.log(data_dict[line][calibrator_name]["INTEN"]), 
                                               data_dict[line][calibrator_name]["AIRMASS"], 
                                               alpha = alpha)

                transmissions[line][calibrator_name]["SLOPE_THEIL"] = fit_result[0]
                transmissions[line][calibrator_name]["INTERCEPT_THEIL"] = fit_result[1]
                transmissions[line][calibrator_name]["LO_SLOPE_THEIL"] = fit_result[2]
                transmissions[line][calibrator_name]["UP_SLOPE_THEIL"] = fit_result[3]
                
                if hasattr(stats, "siegelslopes"):
                    siegel_result = stats.siegelslopes(np.log(data_dict[line][calibrator_name]["INTEN"]), 
                                                   data_dict[line][calibrator_name]["AIRMASS"])
                else:
                    logging.warning("Installed version of scipy does not have the siegelslopes method in scipy.stats!")
                    siegel_result = np.array([np.nan, np.nan])

                
                transmissions[line][calibrator_name]["SLOPE_SIEGEL"] = siegel_result[0]
                transmissions[line][calibrator_name]["INTERCEPT_SIEGEL"] = siegel_result[1]

                if plot:
                    fig = plt.figure()
                    ax = fig.add_subplot(111)
                    ax.scatter(data_dict[line][calibrator_name]["AIRMASS"], 
                               np.log(data_dict[line][calibrator_name]["INTEN"]), 
                               color = 'k', label = "Observations")
                    xlim = ax.get_xlim()
                    ylim = ax.get_ylim()
                    slope_line_x = np.linspace(xlim[0], xlim[1], 10)
                    slope_line_y = fit_result[1] + fit_result[0] * slope_line_x
                    slope_line_y_siegel = siegel_result[1] + siegel_result[0] * slope_line_x
                    ax.plot(slope_line_x, 
                            slope_line_y, 
                            lw = 2, 
                            color = 'r', 
                            ls = "-", 
                            label = "Theil Slope = {:.3f}".format(fit_result[0]))
                    ax.plot(slope_line_x, 
                            slope_line_y_siegel, 
                            lw = 2, 
                            color = 'b', 
                            ls = "--", 
                            label = "Siegel Slope = {0:.3f}".format(siegel_result[0]))
                    
                    
                    ax.set_xlabel("Airmass", fontsize = 12)
                    ax.set_ylabel(r"$/ln$(Intensity/R)", fontsize =12)
                    ax.set_xlim(xlim)
                    ax.set_ylim(ylim)
        
                    ax.legend(fontsize = 12)
                    ax.set_title("{} {}".format(line, calibrator_name), fontsize = 12)
            
    return transmissions
    
    


import pytest
import matplotlib.pyplot as plt
import astropy.units as u
import numpy as np
from ..skySurvey import SkySurvey 
from unittest.mock import Mock

# Set up the random number generator.
np.random.seed(1234)

# Load survey
survey = SkySurvey()

BASELINE_DIR = 'baseline'

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_event():
	"""
	Test click map with sample click event
	"""
	fig = plt.figure()
	ax = fig.add_subplot(111)
	click_map = survey.click_map(fig = fig, image_ax = ax)
	event = Mock()
	event.button = 1
	event.inaxes = ax
	event.xdata = 30
	event.ydata = 5
	click_map.on_click(event)
	return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_event_outside():
	"""
	Test click map with sample click event
	"""
	fig = plt.figure()
	ax = fig.add_subplot(111)
	click_map = survey.click_map(fig = fig, image_ax = ax)
	event = Mock()
	event.button = 1
	event.inaxes = None
	event.xdata = 30
	event.ydata = 5
	click_map.on_click(event)
	return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_no_fig():
	"""
	Test init without fig
	"""
	click_map = survey.click_map()
	return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_no_imax():
	"""
	Test init without fig
	"""
	fig = plt.figure()
	click_map = survey.click_map(fig = fig)
	return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_no_specax():
	"""
	Test init without fig
	"""
	fig = plt.figure()
	ax = fig.add_subplot(111)
	click_map = survey.click_map(fig = fig, image_ax = ax)
	return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_projection():
	import cartopy.crs as ccrs
	"""
	Test init without fig
	"""
	fig = plt.figure()
	ax = fig.add_subplot(111, projection = ccrs.Mollweide())
	click_map = survey.click_map(fig = fig, image_ax = ax)
	event = Mock()
	event.button = 1
	event.inaxes = ax
	event.xdata = 30
	event.ydata = 5
	click_map.on_click(event)
	return plt.gcf()

@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance = 20)
def test_basic_click_projection_kwarg():
	import cartopy.crs as ccrs
	"""
	Test init without fig
	"""
	fig = plt.figure()
	click_map = survey.click_map(fig = fig, projection = ccrs.Mollweide())
	return plt.gcf()


import pytest
import numpy as np

from ..dataReduction import compute_transmissions
from ..skySurvey import directory


# Set up the random number generator
np.random.seed(1234)

import os.path
cal_directory = os.path.join(directory, "tests/test_data/test_dataReduction")

def test_basic():
	"""
	Basic transmission Calculation test
	"""

	transmissions_1 = compute_transmissions(cal_directory)
	transmissions_2 = compute_transmissions(cal_directory, 
		lines = ["ha", "nii"], calibrator = ["G300_0"], plot = True)

	assert np.allclose(transmissions_1["ha"]["g300_0"]["SLOPE_THEIL"], 
		transmissions_2["ha"]["G300_0"]["SLOPE_THEIL"])

def test_string_inputs():
	"""
	Basic transmission Calculation test with string input
	"""

	transmissions_1 = compute_transmissions(cal_directory)
	transmissions_2 = compute_transmissions(cal_directory, 
		lines = "ha", calibrator = "G300_0", plot = True)

	assert np.allclose(transmissions_1["ha"]["g300_0"]["SLOPE_THEIL"], 
		transmissions_2["ha"]["G300_0"]["SLOPE_THEIL"])

def test_type_errors():
	"""
	Ensure errors are raised when lines / calibrators not iterable
	"""

	try:
		transmissions = compute_transmissions(cal_directory, lines = 3.0)
	except TypeError:
		try:
			transmissions = compute_transmissions(cal_directory, calibrator = 300.0)
		except TypeError:
			assert True
		else:
			assert False
	else:
		assert False

def test_read_file_list():
	"""
	Check file_list format
	"""
	from ..skySurvey import SkySurvey
	file_list = os.path.join(cal_directory, "ha/combo/g300_0-000814.fts")

	data = SkySurvey(file_list = file_list)
	data2 = SkySurvey(file_list = [file_list])
	assert np.allclose(data["INTEN"], data2["INTEN"])

def test_read_file_invalid():
	"""
	Rejects invalid file_list format
	"""
	from ..skySurvey import SkySurvey
	file_list = 0
	try:
		SkySurvey(file_list = file_list)
	except TypeError:
		assert True
	else:
		assert False

def test_no_extension():
	"""
	Rejects when extension not found
	"""
	from ..skySurvey import SkySurvey
	file_list = os.path.join(cal_directory, "ha/combo/g300_0-000814.fts")
	try:
		SkySurvey(file_list = file_list, extension = "FAKE")
	except ValueError:
		assert True
	else:
		assert False





import pytest
import astropy.units as u
import numpy as np
from ..skySurvey import SkySurvey 

# Set up the random number generator.
np.random.seed(1234)

# Load global survey
survey = SkySurvey()

def test_get_spectrum():
	from astropy.coordinates import SkyCoord
	"""
	basic extraction test
	"""
	l = np.random.random() * 360.
	b = np.random.random() * 180. - 90
	coordinate = SkyCoord(l = l*u.deg, b = b*u.deg, frame = 'galactic')
	spec = survey.get_spectrum(coordinate)
	index = survey.get_spectrum(coordinate, index = True)

	assert np.allclose(spec["DATA"],survey[index]["DATA"], equal_nan = True)

def test_get_spectrum_quantity():
	"""
	ensure not using SkyCoord works
	"""
	l = np.random.random() * 360.
	b = np.random.random() * 180. - 90
	coordinate = [l,b] * u.deg
	coordinate2 = coordinate.value
	spec = survey.get_spectrum(coordinate)
	index = survey.get_spectrum(coordinate2, index = True)

	assert np.allclose(spec["DATA"],survey[index]["DATA"], equal_nan = True)


import pytest
import numpy as np 
import astropy.units as u
from ..skySurvey import SkySurvey 

from ..skySurvey import directory

# Set up the random number generator.
np.random.seed(1234)

# When running tests locally: filename = "/Users/dk/Data/WHAM/wham-ss-DR1-v161116-170912.fits"

def test_remote_load():
    """
    Ensure survey loads from default remote link
    """

    survey = SkySurvey(mode = "remote")
    assert survey["VELOCITY"].unit == u.km/u.s

survey = SkySurvey()

def test_idlsav_load():
    import os.path
    """
    Ensure IDL Save File loading also works
    """
    filename = os.path.join(directory, "data/wham-ss-DR1-v161116-170912.sav")
    survey_idl = SkySurvey(filename = filename)
    assert survey["DATA"].unit == u.R * u.s / u.km

def test_idlsav_load_varerror():
    import os.path
    """
    Ensure IDL Save File loading fails if WHAM data not there
    """
    filename = os.path.join(directory, "data/wham-ss-DR1-v161116-170912.sav")
    try:
        surveey_idl = SkySurvey(filename = filename, idl_var = "test_fail")
    except TypeError:
        assert True
    else:
        assert False

# def test_idlsav_load_nointen():
#         import os.path
#     """
#     Ensure IDL Save File loading also works for developer versions that
#     don't have a "INTEN" field

#     Note: This test is not yet implemented - need a test IDL Save File
#     """
#     cur_directory = os.path.dirname(__file__)
#     filename = os.path.join(cur_directory, "test_data/test_no_inten.sav")
#     survey_idl = SkySurvey(filename = filename)
#     assert survey["INTEN"].unit == u.R
    

def test_section_circle():
    from astropy.coordinates import Angle
    """
    Ensure survey section extraction works
    """

    l = Angle(np.random.random(1) * 360.*u.deg).wrap_at("180d")
    b = Angle((np.random.random(1) * 180.*u.deg) - 90*u.deg)

    center = [l.value,b.value]
    radius = np.random.random()*30.

    circle = survey.sky_section(center, radius)

    assert circle["DATA"].unit == u.R / u.km * u.s

def test_section_circle_coord_radius_number():
    from astropy.coordinates import Angle
    from astropy.coordinates import SkyCoord
    """
    Ensure survey section extraction works
    """

    l = Angle(np.random.random(1) * 360.*u.deg).wrap_at("180d")
    b = Angle((np.random.random(1) * 180.*u.deg) - 90*u.deg)

    center = SkyCoord(l = l, b = b, frame = 'galactic')
    radius = np.random.random()*30.
    if radius < 5:
      radius = 5

    circle = survey.sky_section(center, radius)

    assert circle["DATA"].unit == u.R / u.km * u.s

def test_section_rect_coord():
    from astropy.coordinates import Angle
    from astropy.coordinates import SkyCoord
    """
    Ensure survey section extraction works
    """

    l = Angle(np.random.random(2) * 360.*u.deg).wrap_at("180d")
    b = Angle((np.random.random(2) * 180.*u.deg) - 90*u.deg)

    bounds = SkyCoord(l = l, b = b, frame = 'galactic')

    rect = survey.sky_section(bounds)

    assert rect["DATA"].unit == u.R / u.km * u.s

def test_section_rect():
    from astropy.coordinates import Angle
    from astropy.coordinates import SkyCoord
    """
    Ensure survey section extraction works
    """

    l = Angle(np.random.random(2) * 360.*u.deg).wrap_at("180d")
    b = Angle((np.random.random(2) * 180.*u.deg) - 90*u.deg)

    bounds = [l.min().value, l.max().value, b.min().value, b.max().value]

    rect = survey.sky_section(bounds)

    assert rect["DATA"].unit == u.R / u.km * u.s

def test_section_no_wrap():
    from astropy.coordinates import Angle
    from astropy.coordinates import SkyCoord
    """
    Ensure survey section extraction works
    """

    l = Angle(np.random.random(2) * 360.*u.deg).wrap_at("360d")
    b = Angle((np.random.random(2) * 180.*u.deg) - 90*u.deg)

    bounds = [l.min().value, l.max().value, b.min().value, b.max().value]

    rect = survey.sky_section(bounds, wrap_at_180 = False)

    assert rect["DATA"].unit == u.R / u.km * u.s

def test_bounds_error():
    try:
        survey.sky_section(np.random.random(5))
    except TypeError:
        assert True
    else:
        assert False

def test_no_radius():
    try:
        survey.sky_section(np.random.random(1), radius = None)
    except TypeError:
        assert True
    else:
        assert False

def test_no_radius_coord():
    from astropy.coordinates import SkyCoord
    try:
        survey.sky_section(SkyCoord(l = np.random.random(1) * u.deg, b = np.random.random(1) * u.deg, 
                                    frame = 'galactic'), 
                            radius = None)
    except TypeError:
        assert True
    else:
        assert False

def test_no_radius_len2():
    try:
        survey.sky_section(np.random.random(2), 
                            radius = None)
    except TypeError:
        assert True
    else:
        assert False






import pytest
import numpy as np 
import astropy.units as u
from ..skySurvey import SkySurvey 

survey = SkySurvey()

def test_moment_0_err():
	"""
	zeroth order moment error
	"""
	moment, err = survey.moment(return_sigma = True)
	assert moment.unit == err.unit

def test_moment_0():
	"""
	zeroth order moment
	"""
	moment, err = survey.moment(return_sigma = True)
	moment2 = survey.moment()
	assert np.allclose(moment.value, moment2.value)

def test_moment_1_err():
	"""
	first order moment error
	"""
	moment, err = survey.moment(order = 1, return_sigma = True)
	assert moment.unit == err.unit

def test_moment_1():
	"""
	zeroth order moment
	"""
	moment, err = survey.moment(order = 1, return_sigma = True)
	moment2 = survey.moment(order = 1)
	assert np.allclose(moment.value, moment2.value)

def test_moment_2_err():
	"""
	second order moment error
	"""
	moment, err = survey.moment(order = 2, return_sigma = True)
	assert moment.unit == err.unit

def test_moment_2():
	"""
	zeroth order moment
	"""
	moment, err = survey.moment(order = 2, return_sigma = True)
	moment2 = survey.moment(order = 2)
	assert np.allclose(moment.value, moment2.value)

def test_moment_2_vel_range():
	"""
	second order moment
	"""
	vmin = np.random.random(1) * -150.
	vmax = np.random.random(1) * 150.
	moment, err = survey.moment(order = 2, return_sigma = True, vmin = vmin, vmax = vmax)
	assert moment.unit == err.unit

def test_moment_2_vel_range_unit():
	"""
	second order moment
	"""
	vmin = np.random.random(1) * -150. * u.km/u.s
	vmax = np.random.random(1) * 150. * u.km/u.s
	moment, err = survey.moment(order = 2, return_sigma = True, vmin = vmin, vmax = vmax)
	assert moment.unit == err.unit


import pytest
import numpy as np
from ..skySurvey import SkySurvey

# Set up the random number generator
np.random.seed(1234)




def test_basic():
	"""
	basic test of spectral stacking
	"""
	import astropy.units as u
	survey = SkySurvey()
	survey2 = SkySurvey()
	survey.get_spectral_slab(-50 * u.km/u.s, 50 * u.km/u.s)
	survey2.get_spectral_slab(-50, 50)
	assert np.allclose(survey["DATA"], survey2["DATA"], equal_nan = True)

def test_no_slab():
	"""
	test no region in velocity range
	"""
	survey = SkySurvey()
	try:
		survey.get_spectral_slab(-300,-250)
	except ValueError:
		assert True
	else:
		assert False